# Copyright (C) 2006-2012, 2016 Canonical Ltd
# Authors: Aaron Bentley
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from io import BytesIO

from ... import branch, merge_directive, tests
from ...bzr.bundle import serializer
from ...controldir import ControlDir
from ...transport import memory
from .. import scenarios

load_tests = scenarios.load_tests_apply_scenarios


class TestSendMixin:
    _default_command = ["send", "-o-"]
    _default_wd = "branch"

    def run_send(self, args, cmd=None, rc=0, wd=None, err_re=None):
        if cmd is None:
            cmd = self._default_command
        if wd is None:
            wd = self._default_wd
        if err_re is None:
            err_re = []
        return self.run_bzr(
            cmd + args, retcode=rc, working_dir=wd, error_regexes=err_re
        )

    def get_MD(self, args, cmd=None, wd="branch"):
        md = self.run_send(args, cmd=cmd, wd=wd)[0]
        out = BytesIO(md.encode("utf-8"))
        return merge_directive.MergeDirective.from_lines(out)

    def assertBundleContains(self, revs, args, cmd=None, wd="branch"):
        md = self.get_MD(args, cmd=cmd, wd=wd)
        br = serializer.read_bundle(BytesIO(md.get_raw_bundle()))
        self.assertEqual(set(revs), {r.revision_id for r in br.revisions})


class TestSend(tests.TestCaseWithTransport, TestSendMixin):
    def setUp(self):
        super().setUp()
        grandparent_tree = ControlDir.create_standalone_workingtree("grandparent")
        self.build_tree_contents([("grandparent/file1", b"grandparent")])
        grandparent_tree.add("file1")
        grandparent_tree.commit("initial commit", rev_id=b"rev1")

        parent_bzrdir = grandparent_tree.controldir.sprout("parent")
        parent_tree = parent_bzrdir.open_workingtree()
        parent_tree.commit("next commit", rev_id=b"rev2")

        branch_tree = parent_tree.controldir.sprout("branch").open_workingtree()
        self.build_tree_contents([("branch/file1", b"branch")])
        branch_tree.commit("last commit", rev_id=b"rev3")

    def assertFormatIs(self, fmt_string, md):
        self.assertEqual(fmt_string, md.get_raw_bundle().splitlines()[0])

    def test_uses_parent(self):
        """Parent location is used as a basis by default."""
        errmsg = self.run_send([], rc=3, wd="grandparent")[1]
        self.assertContainsRe(errmsg, "No submit branch known or specified")
        stdout, stderr = self.run_send([])
        self.assertEqual(stderr.count("Using saved parent location"), 1)
        self.assertBundleContains([b"rev3"], [])

    def test_bundle(self):
        """Bundle works like send, except -o is not required."""
        errmsg = self.run_send([], cmd=["bundle"], rc=3, wd="grandparent")[1]
        self.assertContainsRe(errmsg, "No submit branch known or specified")
        stdout, stderr = self.run_send([], cmd=["bundle"])
        self.assertEqual(stderr.count("Using saved parent location"), 1)
        self.assertBundleContains([b"rev3"], [], cmd=["bundle"])

    def test_uses_submit(self):
        """Submit location can be used and set."""
        self.assertBundleContains([b"rev3"], [])
        self.assertBundleContains([b"rev3", b"rev2"], ["../grandparent"])
        # submit location should be auto-remembered
        self.assertBundleContains([b"rev3", b"rev2"], [])

        self.run_send(["../parent"])
        # We still point to ../grandparent
        self.assertBundleContains([b"rev3", b"rev2"], [])
        # Remember parent now
        self.run_send(["../parent", "--remember"])
        # Now we point to parent
        self.assertBundleContains([b"rev3"], [])

        err = self.run_send(["--remember"], rc=3)[1]
        self.assertContainsRe(err, "--remember requires a branch to be specified.")

    def test_revision_branch_interaction(self):
        self.assertBundleContains([b"rev3", b"rev2"], ["../grandparent"])
        self.assertBundleContains([b"rev2"], ["../grandparent", "-r-2"])
        self.assertBundleContains([b"rev3", b"rev2"], ["../grandparent", "-r-2..-1"])
        md = self.get_MD(["-r-2..-1"])
        self.assertEqual(b"rev2", md.base_revision_id)
        self.assertEqual(b"rev3", md.revision_id)

    def test_output(self):
        # check output for consistency
        # win32 stdout converts LF to CRLF,
        # which would break patch-based bundles
        self.assertBundleContains([b"rev3"], [])

    def test_no_common_ancestor(self):
        foo = self.make_branch_and_tree("foo")
        foo.commit("rev a")
        bar = self.make_branch_and_tree("bar")
        bar.commit("rev b")
        self.run_send(["--from", "foo", "../bar"], wd="foo")

    def test_content_options(self):
        """--no-patch and --no-bundle should work and be independant."""
        md = self.get_MD([])
        self.assertIsNot(None, md.bundle)
        self.assertIsNot(None, md.patch)

        md = self.get_MD(["--format=0.9"])
        self.assertIsNot(None, md.bundle)
        self.assertIsNot(None, md.patch)

        md = self.get_MD(["--no-patch"])
        self.assertIsNot(None, md.bundle)
        self.assertIs(None, md.patch)
        self.run_bzr_error(
            ["Format 0.9 does not permit bundle with no patch"],
            ["send", "--no-patch", "--format=0.9", "-o-"],
            working_dir="branch",
        )
        md = self.get_MD(["--no-bundle", ".", "."])
        self.assertIs(None, md.bundle)
        self.assertIsNot(None, md.patch)

        md = self.get_MD(["--no-bundle", "--format=0.9", "../parent", "."])
        self.assertIs(None, md.bundle)
        self.assertIsNot(None, md.patch)

        md = self.get_MD(["--no-bundle", "--no-patch", ".", "."])
        self.assertIs(None, md.bundle)
        self.assertIs(None, md.patch)

        md = self.get_MD(
            ["--no-bundle", "--no-patch", "--format=0.9", "../parent", "."]
        )
        self.assertIs(None, md.bundle)
        self.assertIs(None, md.patch)

    def test_from_option(self):
        self.run_bzr("send", retcode=3)
        md = self.get_MD(["--from", "branch"])
        self.assertEqual(b"rev3", md.revision_id)
        md = self.get_MD(["-f", "branch"])
        self.assertEqual(b"rev3", md.revision_id)

    def test_output_option(self):
        stdout = self.run_bzr("send -f branch --output file1")[0]
        self.assertEqual("", stdout)
        md_file = open("file1", "rb")
        self.addCleanup(md_file.close)
        self.assertContainsRe(md_file.read(), b"rev3")
        stdout = self.run_bzr("send -f branch --output -")[0]
        self.assertContainsRe(stdout, "rev3")

    def test_note_revisions(self):
        stderr = self.run_send([])[1]
        self.assertEndsWith(stderr, "\nBundling 1 revision.\n")

    def test_mailto_option(self):
        b = branch.Branch.open("branch")
        b.get_config_stack().set("mail_client", "editor")
        self.run_bzr_error(
            ("No mail-to address \\(--mail-to\\) or output \\(-o\\) specified",),
            "send -f branch",
        )
        b.get_config_stack().set("mail_client", "bogus")
        self.run_send([])
        self.run_bzr_error(
            ('Bad value "bogus" for option "mail_client"',),
            "send -f branch --mail-to jrandom@example.org",
        )
        b.get_config_stack().set("submit_to", "jrandom@example.org")
        self.run_bzr_error(
            ('Bad value "bogus" for option "mail_client"',), "send -f branch"
        )

    def test_mailto_child_option(self):
        """Make sure that child_submit_to is used."""
        b = branch.Branch.open("branch")
        b.get_config_stack().set("mail_client", "bogus")
        parent = branch.Branch.open("parent")
        parent.get_config_stack().set("child_submit_to", "somebody@example.org")
        self.run_bzr_error(
            ('Bad value "bogus" for option "mail_client"',), "send -f branch"
        )

    def test_format(self):
        md = self.get_MD(["--format=4"])
        self.assertIs(merge_directive.MergeDirective2, md.__class__)
        self.assertFormatIs(b"# Bazaar revision bundle v4", md)

        md = self.get_MD(["--format=0.9"])
        self.assertFormatIs(b"# Bazaar revision bundle v0.9", md)

        md = self.get_MD(["--format=0.9"], cmd=["bundle"])
        self.assertFormatIs(b"# Bazaar revision bundle v0.9", md)
        self.assertIs(merge_directive.MergeDirective, md.__class__)

        self.run_bzr_error(
            ["Bad value .* for option .format."], "send -f branch -o- --format=0.999"
        )[0]

    def test_format_child_option(self):
        br = branch.Branch.open("parent")
        conf = br.get_config_stack()
        conf.set("child_submit_format", "4")
        md = self.get_MD([])
        self.assertIs(merge_directive.MergeDirective2, md.__class__)

        conf.set("child_submit_format", "0.9")
        md = self.get_MD([])
        self.assertFormatIs(b"# Bazaar revision bundle v0.9", md)

        md = self.get_MD([], cmd=["bundle"])
        self.assertFormatIs(b"# Bazaar revision bundle v0.9", md)
        self.assertIs(merge_directive.MergeDirective, md.__class__)

        conf.set("child_submit_format", "0.999")
        self.run_bzr_error(["No such send format '0.999'"], "send -f branch -o-")[0]

    def test_message_option(self):
        self.run_bzr("send", retcode=3)
        md = self.get_MD([])
        self.assertIs(None, md.message)
        md = self.get_MD(["-m", "my message"])
        self.assertEqual("my message", md.message)

    def test_omitted_revision(self):
        md = self.get_MD(["-r-2.."])
        self.assertEqual(b"rev2", md.base_revision_id)
        self.assertEqual(b"rev3", md.revision_id)
        md = self.get_MD(["-r..3", "--from", "branch", "grandparent"], wd=".")
        self.assertEqual(b"rev1", md.base_revision_id)
        self.assertEqual(b"rev3", md.revision_id)

    def test_nonexistant_branch(self):
        self.vfs_transport_factory = memory.MemoryServer
        location = self.get_url("absentdir/")
        out, err = self.run_bzr(["send", "--from", location], retcode=3)
        self.assertEqual(out, "")
        self.assertEqual(err, 'brz: ERROR: Not a branch: "{}".\n'.format(location))


class TestSendStrictMixin(TestSendMixin):
    def make_parent_and_local_branches(self):
        # Create a 'parent' branch as the base
        self.parent_tree = ControlDir.create_standalone_workingtree("parent")
        self.build_tree_contents([("parent/file", b"parent")])
        self.parent_tree.add("file")
        parent = self.parent_tree.commit("first commit")
        # Branch 'local' from parent and do a change
        local_bzrdir = self.parent_tree.controldir.sprout("local")
        self.local_tree = local_bzrdir.open_workingtree()
        self.build_tree_contents([("local/file", b"local")])
        local = self.local_tree.commit("second commit")
        return parent, local

    _default_command = ["send", "-o-", "../parent"]
    _default_wd = "local"
    _default_sent_revs = None
    _default_errors = [
        'Working tree ".*/local/" has uncommitted changes \\(See brz status\\)\\.',
    ]
    _default_additional_error = "Use --no-strict to force the send.\n"
    _default_additional_warning = "Uncommitted changes will not be sent."

    def set_config_send_strict(self, value):
        br = branch.Branch.open("local")
        br.get_config_stack().set("send_strict", value)

    def assertSendFails(self, args):
        out, err = self.run_send(args, rc=3, err_re=self._default_errors)
        self.assertContainsRe(err, self._default_additional_error)

    def assertSendSucceeds(self, args, revs=None, with_warning=False):
        if with_warning:
            err_re = self._default_errors
        else:
            err_re = []
        if revs is None:
            revs = self._default_sent_revs or [self.local]
        out, err = self.run_send(args, err_re=err_re)
        if len(revs) == 1:
            bundling_revs = f"Bundling {len(revs)} revision.\n"
        else:
            bundling_revs = f"Bundling {len(revs)} revisions.\n"
        if with_warning:
            self.assertContainsRe(err, self._default_additional_warning)
            self.assertEndsWith(err, bundling_revs)
        else:
            self.assertEqual(bundling_revs, err)
        md = merge_directive.MergeDirective.from_lines(BytesIO(out.encode("utf-8")))
        self.assertEqual(self.parent, md.base_revision_id)
        br = serializer.read_bundle(BytesIO(md.get_raw_bundle()))
        self.assertEqual(set(revs), {r.revision_id for r in br.revisions})


class TestSendStrictWithoutChanges(tests.TestCaseWithTransport, TestSendStrictMixin):
    def setUp(self):
        super().setUp()
        self.parent, self.local = self.make_parent_and_local_branches()

    def test_send_without_workingtree(self):
        ControlDir.open("local").destroy_workingtree()
        self.assertSendSucceeds([])

    def test_send_default(self):
        self.assertSendSucceeds([])

    def test_send_strict(self):
        self.assertSendSucceeds(["--strict"])

    def test_send_no_strict(self):
        self.assertSendSucceeds(["--no-strict"])

    def test_send_config_var_strict(self):
        self.set_config_send_strict("true")
        self.assertSendSucceeds([])

    def test_send_config_var_no_strict(self):
        self.set_config_send_strict("false")
        self.assertSendSucceeds([])


class TestSendStrictWithChanges(tests.TestCaseWithTransport, TestSendStrictMixin):
    # These are textually the same as test_push.strict_push_change_scenarios,
    # but since the functions are reimplemented here, the definitions are left
    # here too.
    scenarios = [
        ("uncommitted", {"_changes_type": "_uncommitted_changes"}),
        ("pending_merges", {"_changes_type": "_pending_merges"}),
        ("out-of-sync-trees", {"_changes_type": "_out_of_sync_trees"}),
    ]

    _changes_type = None  # Set by load_tests

    def setUp(self):
        super().setUp()
        # load tests set _changes_types to the name of the method we want to
        # call now
        do_changes_func = getattr(self, self._changes_type)
        do_changes_func()

    def _uncommitted_changes(self):
        self.parent, self.local = self.make_parent_and_local_branches()
        # Make a change without committing it
        self.build_tree_contents([("local/file", b"modified")])

    def _pending_merges(self):
        self.parent, self.local = self.make_parent_and_local_branches()
        # Create 'other' branch containing a new file
        other_bzrdir = self.parent_tree.controldir.sprout("other")
        other_tree = other_bzrdir.open_workingtree()
        self.build_tree_contents([("other/other-file", b"other")])
        other_tree.add("other-file")
        other_tree.commit("other commit", rev_id=b"other")
        # Merge and revert, leaving a pending merge
        self.local_tree.merge_from_branch(other_tree.branch)
        self.local_tree.revert(filenames=["other-file"], backups=False)

    def _out_of_sync_trees(self):
        self.parent, self.local = self.make_parent_and_local_branches()
        self.run_bzr(["checkout", "--lightweight", "local", "checkout"])
        # Make a change and commit it
        self.build_tree_contents([("local/file", b"modified in local")])
        self.local_tree.commit("modify file", rev_id=b"modified-in-local")
        # Exercise commands from the checkout directory
        self._default_wd = "checkout"
        self._default_errors = [
            "Working tree is out of date, please run 'brz update'\\.",
        ]
        self._default_sent_revs = [b"modified-in-local", self.local]

    def test_send_default(self):
        self.assertSendSucceeds([], with_warning=True)

    def test_send_with_revision(self):
        self.assertSendSucceeds(
            ["-r", "revid:" + self.local.decode("utf-8")], revs=[self.local]
        )

    def test_send_no_strict(self):
        self.assertSendSucceeds(["--no-strict"])

    def test_send_strict_with_changes(self):
        self.assertSendFails(["--strict"])

    def test_send_respect_config_var_strict(self):
        self.set_config_send_strict("true")
        self.assertSendFails([])
        self.assertSendSucceeds(["--no-strict"])

    def test_send_bogus_config_var_ignored(self):
        self.set_config_send_strict("I'm unsure")
        self.assertSendSucceeds([], with_warning=True)

    def test_send_no_strict_command_line_override_config(self):
        self.set_config_send_strict("true")
        self.assertSendFails([])
        self.assertSendSucceeds(["--no-strict"])

    def test_send_strict_command_line_override_config(self):
        self.set_config_send_strict("false")
        self.assertSendSucceeds([])
        self.assertSendFails(["--strict"])


class TestBundleStrictWithoutChanges(TestSendStrictWithoutChanges):
    _default_command = ["bundle-revisions", "../parent"]



# Copyright (C) 2005-2012, 2016 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
from io import BytesIO

from .. import conflicts, errors, trace, workingtree
from ..bzr import bzrdir, workingtree_3, workingtree_4
from ..bzr import conflicts as _mod_bzr_conflicts
from ..bzr import workingtree as bzrworkingtree
from ..lock import write_locked
from ..lockdir import LockDir
from ..tree import TreeDirectory, TreeEntry, TreeFile, TreeLink
from . import TestCase, TestCaseWithTransport, TestSkipped
from .features import SymlinkFeature


class TestTreeDirectory(TestCaseWithTransport):
    def test_kind_character(self):
        self.assertEqual(TreeDirectory().kind_character(), "/")


class TestTreeEntry(TestCaseWithTransport):
    def test_kind_character(self):
        self.assertEqual(TreeEntry().kind_character(), "???")


class TestTreeFile(TestCaseWithTransport):
    def test_kind_character(self):
        self.assertEqual(TreeFile().kind_character(), "")


class TestTreeLink(TestCaseWithTransport):
    def test_kind_character(self):
        self.assertEqual(TreeLink().kind_character(), "")


class TestDefaultFormat(TestCaseWithTransport):
    def test_get_set_default_format(self):
        old_format = workingtree.format_registry.get_default()
        # default is 6
        self.assertTrue(isinstance(old_format, workingtree_4.WorkingTreeFormat6))
        workingtree.format_registry.set_default(SampleTreeFormat())
        try:
            # the default branch format is used by the meta dir format
            # which is not the default bzrdir format at this point
            dir = bzrdir.BzrDirMetaFormat1().initialize(".")
            dir.create_repository()
            dir.create_branch()
            result = dir.create_workingtree()
            self.assertEqual(result, "A tree")
        finally:
            workingtree.format_registry.set_default(old_format)
        self.assertEqual(old_format, workingtree.format_registry.get_default())

    def test_from_string(self):
        self.assertIsInstance(
            SampleTreeFormat.from_string(b"Sample tree format."), SampleTreeFormat
        )
        self.assertRaises(
            AssertionError, SampleTreeFormat.from_string, b"Different format string."
        )

    def test_get_set_default_format_by_key(self):
        old_format = workingtree.format_registry.get_default()
        # default is 6
        format = SampleTreeFormat()
        workingtree.format_registry.register(format)
        self.addCleanup(workingtree.format_registry.remove, format)
        self.assertTrue(isinstance(old_format, workingtree_4.WorkingTreeFormat6))
        workingtree.format_registry.set_default_key(format.get_format_string())
        try:
            # the default branch format is used by the meta dir format
            # which is not the default bzrdir format at this point
            dir = bzrdir.BzrDirMetaFormat1().initialize(".")
            dir.create_repository()
            dir.create_branch()
            result = dir.create_workingtree()
            self.assertEqual(result, "A tree")
        finally:
            workingtree.format_registry.set_default_key(old_format.get_format_string())
        self.assertEqual(old_format, workingtree.format_registry.get_default())

    def test_open(self):
        tree = self.make_branch_and_tree(".")
        open_direct = workingtree.WorkingTree.open(".")
        self.assertEqual(tree.basedir, open_direct.basedir)
        open_no_args = workingtree.WorkingTree.open()
        self.assertEqual(tree.basedir, open_no_args.basedir)

    def test_open_containing(self):
        tree = self.make_branch_and_tree(".")
        open_direct, relpath = workingtree.WorkingTree.open_containing(".")
        self.assertEqual(tree.basedir, open_direct.basedir)
        self.assertEqual("", relpath)
        open_no_args, relpath = workingtree.WorkingTree.open_containing()
        self.assertEqual(tree.basedir, open_no_args.basedir)
        self.assertEqual("", relpath)
        open_subdir, relpath = workingtree.WorkingTree.open_containing("subdir")
        self.assertEqual(tree.basedir, open_subdir.basedir)
        self.assertEqual("subdir", relpath)


class SampleTreeFormat(bzrworkingtree.WorkingTreeFormatMetaDir):
    """A sample format.

    this format is initializable, unsupported to aid in testing the
    open and open_downlevel routines.
    """

    @classmethod
    def get_format_string(cls):
        """See WorkingTreeFormat.get_format_string()."""
        return b"Sample tree format."

    def initialize(
        self,
        a_controldir,
        revision_id=None,
        from_branch=None,
        accelerator_tree=None,
        hardlink=False,
    ):
        """Sample branches cannot be created."""
        t = a_controldir.get_workingtree_transport(self)
        t.put_bytes("format", self.get_format_string())
        return "A tree"

    def is_supported(self):
        return False

    def open(self, transport, _found=False):
        return "opened tree."


class SampleExtraTreeFormat(workingtree.WorkingTreeFormat):
    """A sample format that does not support use in a metadir."""

    def get_format_string(self):
        # Not usable in a metadir, so no format string
        return None

    def initialize(
        self,
        a_controldir,
        revision_id=None,
        from_branch=None,
        accelerator_tree=None,
        hardlink=False,
    ):
        raise NotImplementedError(self.initialize)

    def is_supported(self):
        return False

    def open(self, transport, _found=False):
        raise NotImplementedError(self.open)


class TestWorkingTreeFormat(TestCaseWithTransport):
    """Tests for the WorkingTreeFormat facility."""

    def test_find_format_string(self):
        # is the right format object found for a working tree?
        branch = self.make_branch("branch")
        self.assertRaises(
            errors.NoWorkingTree,
            bzrworkingtree.WorkingTreeFormatMetaDir.find_format_string,
            branch.controldir,
        )
        transport = branch.controldir.get_workingtree_transport(None)
        transport.mkdir(".")
        transport.put_bytes("format", b"some format name")
        # The format does not have to be known by Bazaar,
        # find_format_string just retrieves the name
        self.assertEqual(
            b"some format name",
            bzrworkingtree.WorkingTreeFormatMetaDir.find_format_string(
                branch.controldir
            ),
        )

    def test_find_format(self):
        # is the right format object found for a working tree?
        # create a branch with a few known format objects.
        self.build_tree(["foo/", "bar/"])

        def check_format(format, url):
            dir = format._matchingcontroldir.initialize(url)
            dir.create_repository()
            dir.create_branch()
            format.initialize(dir)
            found_format = bzrworkingtree.WorkingTreeFormatMetaDir.find_format(dir)
            self.assertIsInstance(found_format, format.__class__)

        check_format(workingtree_3.WorkingTreeFormat3(), "bar")

    def test_find_format_no_tree(self):
        dir = bzrdir.BzrDirMetaFormat1().initialize(".")
        self.assertRaises(
            errors.NoWorkingTree,
            bzrworkingtree.WorkingTreeFormatMetaDir.find_format,
            dir,
        )

    def test_find_format_unknown_format(self):
        dir = bzrdir.BzrDirMetaFormat1().initialize(".")
        dir.create_repository()
        dir.create_branch()
        SampleTreeFormat().initialize(dir)
        self.assertRaises(
            errors.UnknownFormatError,
            bzrworkingtree.WorkingTreeFormatMetaDir.find_format,
            dir,
        )

    def test_find_format_with_features(self):
        tree = self.make_branch_and_tree(".", format="2a")
        tree.update_feature_flags({b"name": b"necessity"})
        found_format = bzrworkingtree.WorkingTreeFormatMetaDir.find_format(
            tree.controldir
        )
        self.assertIsInstance(found_format, workingtree.WorkingTreeFormat)
        self.assertEqual(found_format.features.get(b"name"), b"necessity")
        self.assertRaises(
            bzrdir.MissingFeature, found_format.check_support_status, True
        )
        self.addCleanup(
            bzrworkingtree.WorkingTreeFormatMetaDir.unregister_feature, b"name"
        )
        bzrworkingtree.WorkingTreeFormatMetaDir.register_feature(b"name")
        found_format.check_support_status(True)


class TestWorkingTreeIterEntriesByDir_wSubtrees(TestCaseWithTransport):
    def make_simple_tree(self):
        tree = self.make_branch_and_tree("tree", format="development-subtree")
        self.build_tree(["tree/a/", "tree/a/b/", "tree/a/b/c"])
        tree.set_root_id(b"root-id")
        tree.add(["a", "a/b", "a/b/c"], ids=[b"a-id", b"b-id", b"c-id"])
        tree.commit("initial")
        return tree

    def test_just_directory(self):
        tree = self.make_simple_tree()
        self.assertEqual(
            [
                ("directory", b"root-id"),
                ("directory", b"a-id"),
                ("directory", b"b-id"),
                ("file", b"c-id"),
            ],
            [(ie.kind, ie.file_id) for path, ie in tree.iter_entries_by_dir()],
        )
        self.make_branch_and_tree("tree/a/b")
        self.assertEqual(
            [("tree-reference", b"b-id")],
            [
                (ie.kind, ie.file_id)
                for path, ie in tree.iter_entries_by_dir(specific_files=["a/b"])
            ],
        )

    def test_direct_subtree(self):
        tree = self.make_simple_tree()
        self.make_branch_and_tree("tree/a/b")
        self.assertEqual(
            [
                ("directory", b"root-id"),
                ("directory", b"a-id"),
                ("tree-reference", b"b-id"),
            ],
            [(ie.kind, ie.file_id) for path, ie in tree.iter_entries_by_dir()],
        )

    def test_indirect_subtree(self):
        tree = self.make_simple_tree()
        self.make_branch_and_tree("tree/a")
        self.assertEqual(
            [("directory", b"root-id"), ("tree-reference", b"a-id")],
            [(ie.kind, ie.file_id) for path, ie in tree.iter_entries_by_dir()],
        )


class TestWorkingTreeFormatRegistry(TestCase):
    def setUp(self):
        super().setUp()
        self.registry = workingtree.WorkingTreeFormatRegistry()

    def test_register_unregister_format(self):
        format = SampleTreeFormat()
        self.registry.register(format)
        self.assertEqual(format, self.registry.get(b"Sample tree format."))
        self.registry.remove(format)
        self.assertRaises(KeyError, self.registry.get, b"Sample tree format.")

    def test_get_all(self):
        format = SampleTreeFormat()
        self.assertEqual([], self.registry._get_all())
        self.registry.register(format)
        self.assertEqual([format], self.registry._get_all())

    def test_register_extra(self):
        format = SampleExtraTreeFormat()
        self.assertEqual([], self.registry._get_all())
        self.registry.register_extra(format)
        self.assertEqual([format], self.registry._get_all())

    def test_register_extra_lazy(self):
        self.assertEqual([], self.registry._get_all())
        self.registry.register_extra_lazy(
            "breezy.tests.test_workingtree", "SampleExtraTreeFormat"
        )
        formats = self.registry._get_all()
        self.assertEqual(1, len(formats))
        self.assertIsInstance(formats[0], SampleExtraTreeFormat)


class TestWorkingTreeFormat3(TestCaseWithTransport):
    """Tests specific to WorkingTreeFormat3."""

    def test_disk_layout(self):
        control = bzrdir.BzrDirMetaFormat1().initialize(self.get_url())
        control.create_repository()
        control.create_branch()
        workingtree_3.WorkingTreeFormat3().initialize(control)
        # we want:
        # format 'Bazaar-NG Working Tree format 3'
        # inventory = blank inventory
        # pending-merges = ''
        # stat-cache = ??
        # no inventory.basis yet
        t = control.get_workingtree_transport(None)
        self.assertEqualDiff(b"Bazaar-NG Working Tree format 3", t.get("format").read())
        self.assertEqualDiff(
            t.get("inventory").read(),
            b'<inventory format="5">\n</inventory>\n',
        )
        self.assertEqualDiff(b"### bzr hashcache v5\n", t.get("stat-cache").read())
        self.assertFalse(t.has("inventory.basis"))
        # no last-revision file means 'None' or 'NULLREVISION'
        self.assertFalse(t.has("last-revision"))
        # TODO RBC 20060210 do a commit, check the inventory.basis is created
        # correctly and last-revision file becomes present.

    def test_uses_lockdir(self):
        """WorkingTreeFormat3 uses its own LockDir:

        - lock is a directory
        - when the WorkingTree is locked, LockDir can see that
        """
        t = self.get_transport()
        url = self.get_url()
        dir = bzrdir.BzrDirMetaFormat1().initialize(url)
        dir.create_repository()
        dir.create_branch()
        try:
            tree = workingtree_3.WorkingTreeFormat3().initialize(dir)
        except errors.NotLocalUrl:
            raise TestSkipped("Not a local URL")
        self.assertIsDirectory(".bzr", t)
        self.assertIsDirectory(".bzr/checkout", t)
        self.assertIsDirectory(".bzr/checkout/lock", t)
        our_lock = LockDir(t, ".bzr/checkout/lock")
        self.assertEqual(our_lock.peek(), None)
        with tree.lock_write():
            self.assertTrue(our_lock.peek())
        self.assertEqual(our_lock.peek(), None)

    def test_missing_pending_merges(self):
        control = bzrdir.BzrDirMetaFormat1().initialize(self.get_url())
        control.create_repository()
        control.create_branch()
        tree = workingtree_3.WorkingTreeFormat3().initialize(control)
        tree._transport.delete("pending-merges")
        self.assertEqual([], tree.get_parent_ids())


class TestRevert(TestCaseWithTransport):
    def test_revert_conflicts_recursive(self):
        this_tree = self.make_branch_and_tree("this-tree")
        self.build_tree_contents([("this-tree/foo/",), ("this-tree/foo/bar", b"bar")])
        this_tree.add(["foo", "foo/bar"])
        this_tree.commit("created foo/bar")
        other_tree = this_tree.controldir.sprout("other-tree").open_workingtree()
        self.build_tree_contents([("other-tree/foo/bar", b"baz")])
        other_tree.commit("changed bar")
        self.build_tree_contents([("this-tree/foo/bar", b"qux")])
        this_tree.commit("changed qux")
        this_tree.merge_from_branch(other_tree.branch)
        self.assertEqual(1, len(this_tree.conflicts()))
        this_tree.revert(["foo"])
        self.assertEqual(0, len(this_tree.conflicts()))


class TestAutoResolve(TestCaseWithTransport):
    def test_auto_resolve(self):
        base = self.make_branch_and_tree("base")
        self.build_tree_contents([("base/hello", b"Hello")])
        base.add("hello", ids=b"hello_id")
        base.commit("Hello")
        other = base.controldir.sprout("other").open_workingtree()
        self.build_tree_contents([("other/hello", b"hELLO")])
        other.commit("Case switch")
        this = base.controldir.sprout("this").open_workingtree()
        self.assertPathExists("this/hello")
        self.build_tree_contents([("this/hello", b"Hello World")])
        this.commit("Add World")
        this.merge_from_branch(other.branch)
        self.assertEqual(
            [_mod_bzr_conflicts.TextConflict("hello", b"hello_id")], this.conflicts()
        )
        this.auto_resolve()
        self.assertEqual(
            [_mod_bzr_conflicts.TextConflict("hello", b"hello_id")], this.conflicts()
        )
        self.build_tree_contents([("this/hello", "<<<<<<<")])
        this.auto_resolve()
        self.assertEqual(
            [_mod_bzr_conflicts.TextConflict("hello", b"hello_id")], this.conflicts()
        )
        self.build_tree_contents([("this/hello", "=======")])
        this.auto_resolve()
        self.assertEqual(
            [_mod_bzr_conflicts.TextConflict("hello", b"hello_id")], this.conflicts()
        )
        self.build_tree_contents([("this/hello", "\n>>>>>>>")])
        remaining, resolved = this.auto_resolve()
        self.assertEqual(
            [_mod_bzr_conflicts.TextConflict("hello", b"hello_id")], this.conflicts()
        )
        self.assertEqual([], resolved)
        self.build_tree_contents([("this/hello", b"hELLO wORLD")])
        remaining, resolved = this.auto_resolve()
        self.assertEqual([], this.conflicts())
        self.assertEqual(
            [_mod_bzr_conflicts.TextConflict("hello", b"hello_id")], resolved
        )
        self.assertPathDoesNotExist("this/hello.BASE")

    def test_unsupported_symlink_auto_resolve(self):
        self.requireFeature(SymlinkFeature(self.test_dir))
        base = self.make_branch_and_tree("base")
        self.build_tree_contents([("base/hello", "Hello")])
        base.add("hello", ids=b"hello_id")
        base.commit("commit 0")
        other = base.controldir.sprout("other").open_workingtree()
        self.build_tree_contents([("other/hello", "Hello")])
        os.symlink("other/hello", "other/foo")
        other.add("foo", ids=b"foo_id")
        other.commit("commit symlink")
        this = base.controldir.sprout("this").open_workingtree()
        self.assertPathExists("this/hello")
        self.build_tree_contents([("this/hello", "Hello")])
        this.commit("commit 2")
        log = BytesIO()
        trace.push_log_file(log)
        os_symlink = getattr(os, "symlink", None)
        os.symlink = None
        try:
            this.merge_from_branch(other.branch)
        finally:
            if os_symlink:
                os.symlink = os_symlink
        self.assertContainsRe(
            log.getvalue(), b'Unable to create symlink "foo" on this filesystem'
        )

    def test_auto_resolve_dir(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/hello/"])
        tree.add("hello", ids=b"hello-id")
        file_conflict = _mod_bzr_conflicts.TextConflict("hello", b"hello-id")
        tree.set_conflicts([file_conflict])
        remaining, resolved = tree.auto_resolve()
        self.assertEqual(
            remaining,
            conflicts.ConflictList(
                [_mod_bzr_conflicts.TextConflict("hello", "hello-id")]
            ),
        )
        self.assertEqual(resolved, [])

    def test_auto_resolve_missing(self):
        tree = self.make_branch_and_tree("tree")
        file_conflict = _mod_bzr_conflicts.TextConflict("hello", b"hello-id")
        tree.set_conflicts([file_conflict])
        remaining, resolved = tree.auto_resolve()
        self.assertEqual(remaining, [])
        self.assertEqual(
            resolved,
            conflicts.ConflictList(
                [_mod_bzr_conflicts.TextConflict("hello", "hello-id")]
            ),
        )


class TestStoredUncommitted(TestCaseWithTransport):
    def store_uncommitted(self):
        tree = self.make_branch_and_tree("tree")
        tree.commit("get root in there")
        self.build_tree_contents([("tree/file", b"content")])
        tree.add("file", ids=b"file-id")
        tree.store_uncommitted()
        return tree

    def test_store_uncommitted(self):
        self.store_uncommitted()
        self.assertPathDoesNotExist("tree/file")

    def test_store_uncommitted_no_change(self):
        tree = self.make_branch_and_tree("tree")
        tree.commit("get root in there")
        tree.store_uncommitted()
        self.assertIs(None, tree.branch.get_unshelver(tree))

    def test_restore_uncommitted(self):
        with write_locked(self.store_uncommitted()) as tree:
            tree.restore_uncommitted()
            self.assertPathExists("tree/file")
            self.assertIs(None, tree.branch.get_unshelver(tree))

    def test_restore_uncommitted_none(self):
        tree = self.make_branch_and_tree("tree")
        tree.restore_uncommitted()



# Copyright (C) 2008-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test directory service implementation."""

from .. import transport, urlutils
from ..directory_service import (
    AliasDirectory,
    DirectoryServiceRegistry,
    InvalidLocationAlias,
    UnsetLocationAlias,
    directories,
)
from . import TestCase, TestCaseWithTransport


class FooService:
    """A directory service that maps the name to a FILE url."""

    # eg 'file:///foo' on Unix, or 'file:///C:/foo' on Windows
    base = urlutils.local_path_to_url("/foo")

    def look_up(self, name, url, purpose=None):
        return self.base + name


class TestDirectoryLookup(TestCase):
    def setUp(self):
        super().setUp()
        self.registry = DirectoryServiceRegistry()
        self.registry.register("foo:", FooService, "Map foo URLs to http urls")

    def test_get_directory_service(self):
        directory, suffix = self.registry.get_prefix("foo:bar")
        self.assertIs(FooService, directory)
        self.assertEqual("bar", suffix)

    def test_dereference(self):
        self.assertEqual(FooService.base + "bar", self.registry.dereference("foo:bar"))
        self.assertEqual(
            FooService.base + "bar",
            self.registry.dereference("foo:bar", purpose="write"),
        )
        self.assertEqual("baz:qux", self.registry.dereference("baz:qux"))
        self.assertEqual(
            "baz:qux", self.registry.dereference("baz:qux", purpose="write")
        )

    def test_get_transport(self):
        directories.register("foo:", FooService, "Map foo URLs to http urls")
        self.addCleanup(directories.remove, "foo:")
        self.assertEqual(
            FooService.base + "bar/", transport.get_transport("foo:bar").base
        )


class OldService:
    """A directory service that maps the name to a FILE url."""

    # eg 'file:///foo' on Unix, or 'file:///C:/foo' on Windows
    base = urlutils.local_path_to_url("/foo")

    def look_up(self, name, url):
        return self.base + name


class TestOldDirectoryLookup(TestCase):
    """Test compatibility with older implementations of Directory
    that don't support the purpose argument.
    """

    def setUp(self):
        super().setUp()
        self.registry = DirectoryServiceRegistry()
        self.registry.register("old:", OldService, "Map foo URLs to http urls")

    def test_dereference(self):
        self.assertEqual(OldService.base + "bar", self.registry.dereference("old:bar"))
        self.assertEqual(
            OldService.base + "bar",
            self.registry.dereference("old:bar", purpose="write"),
        )
        self.assertEqual("baz:qux", self.registry.dereference("baz:qux"))
        self.assertEqual(
            "baz:qux", self.registry.dereference("baz:qux", purpose="write")
        )


class TestAliasDirectory(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.branch = self.make_branch(".")

    def assertAliasFromBranch(self, setter, value, alias):
        setter(value)
        self.assertEqual(value, directories.dereference(alias))

    def test_lookup_parent(self):
        self.assertAliasFromBranch(self.branch.set_parent, "http://a", ":parent")

    def test_lookup_submit(self):
        self.assertAliasFromBranch(self.branch.set_submit_branch, "http://b", ":submit")

    def test_lookup_public(self):
        self.assertAliasFromBranch(self.branch.set_public_branch, "http://c", ":public")

    def test_lookup_bound(self):
        self.assertAliasFromBranch(self.branch.set_bound_location, "http://d", ":bound")

    def test_lookup_push(self):
        self.assertAliasFromBranch(self.branch.set_push_location, "http://e", ":push")

    def test_lookup_this(self):
        self.assertEqual(self.branch.base, directories.dereference(":this"))

    def test_extra_path(self):
        self.assertEqual(
            urlutils.join(self.branch.base, "arg"), directories.dereference(":this/arg")
        )

    def test_lookup_badname(self):
        e = self.assertRaises(InvalidLocationAlias, directories.dereference, ":booga")
        self.assertEqual('":booga" is not a valid location alias.', str(e))

    def test_lookup_badvalue(self):
        e = self.assertRaises(UnsetLocationAlias, directories.dereference, ":parent")
        self.assertEqual("No parent location assigned.", str(e))

    def test_register_location_alias(self):
        self.addCleanup(AliasDirectory.branch_aliases.remove, "booga")
        AliasDirectory.branch_aliases.register(
            "booga", lambda b: "UHH?", help="Nobody knows"
        )
        self.assertEqual("UHH?", directories.dereference(":booga"))


class TestColocatedDirectory(TestCaseWithTransport):
    def test_lookup_non_default(self):
        default = self.make_branch(".")
        non_default = default.controldir.create_branch(name="nondefault")
        self.assertEqual(non_default.base, directories.dereference("co:nondefault"))

    def test_lookup_default(self):
        default = self.make_branch(".")
        default.controldir.create_branch(name="nondefault")
        self.assertEqual(
            urlutils.join_segment_parameters(
                default.controldir.user_url, {"branch": ""}
            ),
            directories.dereference("co:"),
        )

    def test_no_such_branch(self):
        # No error is raised in this case, that is up to the code that actually
        # opens the branch.
        default = self.make_branch(".")
        self.assertEqual(
            urlutils.join_segment_parameters(
                default.controldir.user_url, {"branch": "foo"}
            ),
            directories.dereference("co:foo"),
        )



# Copyright (C) 2006, 2007, 2009-2012, 2016 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy import tests
from breezy.revision import NULL_REVISION
from breezy.tests import per_workingtree


class TestPull(per_workingtree.TestCaseWithWorkingTree):
    def get_pullable_trees(self):
        self.build_tree(["from/", "from/file", "to/"])
        tree = self.make_branch_and_tree("from")
        tree.add("file")
        a = tree.commit("foo")
        tree_b = self.make_branch_and_tree("to")
        return tree, tree_b, a

    def test_pull_null(self):
        tree_a, tree_b, rev_a = self.get_pullable_trees()
        if tree_a.supports_setting_file_ids():
            root_id = tree_a.path2id("")
            tree_a.pull(tree_b.branch, stop_revision=NULL_REVISION, overwrite=True)
            self.assertEqual(root_id, tree_a.path2id(""))
        else:
            tree_a.pull(tree_b.branch, stop_revision=NULL_REVISION, overwrite=True)

    def test_pull(self):
        tree_a, tree_b, rev_a = self.get_pullable_trees()
        tree_b.pull(tree_a.branch)
        self.assertTrue(tree_b.branch.repository.has_revision(rev_a))
        self.assertEqual([rev_a], tree_b.get_parent_ids())

    def test_pull_overwrites(self):
        tree_a, tree_b, rev_a = self.get_pullable_trees()
        rev_b = tree_b.commit("foo")
        self.assertEqual(rev_b, tree_b.branch.last_revision())
        tree_b.pull(tree_a.branch, overwrite=True)
        self.assertTrue(tree_b.branch.repository.has_revision(rev_a))
        self.assertTrue(tree_b.branch.repository.has_revision(rev_b))
        self.assertEqual([rev_a], tree_b.get_parent_ids())

    def test_pull_merges_tree_content(self):
        tree_a, tree_b, rev_a = self.get_pullable_trees()
        tree_b.pull(tree_a.branch)
        self.assertFileEqual(b"contents of from/file\n", "to/file")

    def test_pull_changes_root_id(self):
        tree = self.make_branch_and_tree("from")
        if not tree._format.supports_versioned_directories:
            self.skipTest("format does not support custom root ids")
        tree.set_root_id(b"first_root_id")
        self.build_tree(["from/file"])
        tree.add(["file"])
        tree.commit("first")
        to_tree = tree.controldir.sprout("to").open_workingtree()
        self.assertEqual(b"first_root_id", to_tree.path2id(""))
        tree.set_root_id(b"second_root_id")
        tree.commit("second")
        to_tree.pull(tree.branch)
        self.assertEqual(b"second_root_id", to_tree.path2id(""))


class TestPullWithOrphans(per_workingtree.TestCaseWithWorkingTree):
    def make_branch_deleting_dir(self, relpath=None):
        if relpath is None:
            relpath = "trunk"
        builder = self.make_branch_builder(relpath)
        builder.start_series()

        # Create an empty trunk
        builder.build_snapshot(
            None, [("add", ("", b"root-id", "directory", ""))], revision_id=b"1"
        )
        builder.build_snapshot(
            [b"1"],
            [
                ("add", ("dir", b"dir-id", "directory", "")),
                ("add", ("file", b"file-id", "file", b"trunk content\n")),
            ],
            revision_id=b"2",
        )
        builder.build_snapshot(
            [b"2"],
            [
                ("unversion", "dir"),
            ],
            revision_id=b"3",
        )
        builder.finish_series()
        return builder.get_branch()

    def test_pull_orphans(self):
        if not self.workingtree_format.missing_parent_conflicts:
            raise tests.TestSkipped(
                "{!r} does not support missing parent conflicts".format(
                    self.workingtree_format
                )
            )
        trunk = self.make_branch_deleting_dir("trunk")
        work = trunk.controldir.sprout("work", revision_id=b"2").open_workingtree()
        work.branch.get_config_stack().set("transform.orphan_policy", "move")
        # Add some unversioned files in dir
        self.build_tree(["work/dir/foo", "work/dir/subdir/", "work/dir/subdir/foo"])
        work.pull(trunk)
        self.assertLength(0, work.conflicts())
        # The directory removal should succeed
        self.assertPathDoesNotExist("work/dir")



# Copyright (C) 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interface conformance of 'WorkingTree.move'."""

import os

from breezy import errors, osutils, tests
from breezy.tests import features
from breezy.tests.matchers import HasLayout, HasPathRelations
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestMove(TestCaseWithWorkingTree):
    def assertPathRelations(self, previous_tree, tree, relations):
        self.assertThat(tree, HasPathRelations(previous_tree, relations))

    def assertTreeLayout(self, expected, tree):
        """Check that the tree has the correct layout."""
        self.assertThat(tree, HasLayout(expected))

    def test_move_via_rm_and_add(self):
        """Move by remove and add-with-id."""
        self.build_tree(["a1", "b1"])
        tree = self.make_branch_and_tree(".")
        if tree.supports_setting_file_ids():
            tree.add(["a1"], ids=[b"a1-id"])
        else:
            tree.add(["a1"])
        tree.commit("initial commit")
        tree.remove("a1", force=True, keep_files=False)
        if tree.supports_setting_file_ids():
            tree.add(["b1"], ids=[b"a1-id"])
        else:
            tree.add(["b1"])
        tree._validate()

    def test_move_correct_call_named(self):
        """tree.move has the deprecated parameter 'to_name'.
        It has been replaced by 'to_dir' for consistency.
        Test the new API using named parameter.
        """
        self.build_tree(["a1", "sub1/"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a1", "sub1"])
        tree.commit("initial commit")
        self.assertEqual(
            [("a1", "sub1/a1")], tree.move(["a1"], to_dir="sub1", after=False)
        )
        tree._validate()

    def test_move_correct_call_unnamed(self):
        """tree.move has the deprecated parameter 'to_name'.
        It has been replaced by 'to_dir' for consistency.
        Test the new API using unnamed parameter.
        """
        self.build_tree(["a1", "sub1/"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a1", "sub1"])
        tree.commit("initial commit")
        self.assertEqual([("a1", "sub1/a1")], tree.move(["a1"], "sub1", after=False))
        tree._validate()

    def test_move_target_not_dir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        tree.add(["a"])
        tree.commit("initial")

        self.assertRaises(errors.BzrMoveFailedError, tree.move, ["a"], "not-a-dir")
        tree._validate()

    def test_move_non_existent(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/"])
        tree.add(["a"])
        tree.commit("initial")
        self.assertRaises(errors.BzrMoveFailedError, tree.move, ["not-a-file"], "a")
        self.assertRaises(errors.BzrMoveFailedError, tree.move, ["not-a-file"], "")
        tree._validate()

    def test_move_target_not_versioned(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "b"])
        tree.add(["b"])
        tree.commit("initial")
        if tree.has_versioned_directories():
            self.assertRaises(errors.BzrMoveFailedError, tree.move, ["b"], "a")
        else:
            tree.move(["b"], "a")
        tree._validate()

    def test_move_unversioned(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "b"])
        tree.add(["a"])
        tree.commit("initial")
        self.assertRaises(errors.BzrMoveFailedError, tree.move, ["b"], "a")
        tree._validate()

    def test_move_multi_unversioned(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "b", "c", "d"])
        tree.add(["a", "c", "d"])
        tree.commit("initial")
        self.assertRaises(errors.BzrMoveFailedError, tree.move, ["c", "b", "d"], "a")
        self.assertRaises(errors.BzrMoveFailedError, tree.move, ["b", "c", "d"], "a")
        self.assertRaises(errors.BzrMoveFailedError, tree.move, ["d", "c", "b"], "a")
        if osutils.lexists("a/c"):
            # If 'c' was actually moved, then 'd' should have also been moved
            self.assertPathRelations(
                tree.basis_tree(),
                tree,
                [("", ""), ("a/", "a/"), ("a/c", "c"), ("a/d", "d")],
            )
        else:
            self.assertPathRelations(
                tree.basis_tree(),
                tree,
                [("", ""), ("a/", "a/"), ("c", "c"), ("d", "d")],
            )
        tree._validate()

    def test_move_over_deleted(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "b"])
        tree.add(["a", "a/b", "b"])
        tree.commit("initial")

        tree.remove(["a/b"], keep_files=False)
        self.assertEqual([("b", "a/b")], tree.move(["b"], "a"))
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a/", "a/"), ("a/b", "b")]
        )
        tree._validate()

    def test_move_subdir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/c"])
        tree.add(["a", "b", "b/c"])
        tree.commit("initial")
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("a", "a"), ("b/", "b/"), ("b/c", "b/c")],
        )
        a_contents = tree.get_file_text("a")
        self.assertEqual([("a", "b/a")], tree.move(["a"], "b"))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("b/", "b/"), ("b/a", "a"), ("b/c", "b/c")],
        )
        self.assertPathDoesNotExist("a")
        self.assertFileEqual(a_contents, "b/a")
        tree._validate()

    def test_move_parent_dir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/c"])
        tree.add(["a", "b", "b/c"])
        tree.commit("initial")
        c_contents = tree.get_file_text("b/c")
        self.assertEqual([("b/c", "c")], tree.move(["b/c"], ""))
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b/", "b/"), ("c", "b/c")]
        )
        self.assertPathDoesNotExist("b/c")
        self.assertFileEqual(c_contents, "c")
        tree._validate()

    def test_move_fail_consistent(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/a", "c"])
        tree.add(["a", "b", "c"])
        tree.commit("initial")
        # Target already exists
        self.assertRaises(errors.RenameFailedFilesExist, tree.move, ["c", "a"], "b")
        # 'c' may or may not have been moved, but either way the tree should
        # maintain a consistent state.
        if osutils.lexists("c"):
            self.assertPathRelations(
                tree.basis_tree(),
                tree,
                [("", ""), ("a", "a"), ("b/", "b/"), ("c", "c")],
            )
        else:
            self.assertPathExists("b/c")
            self.assertPathRelations(
                tree.basis_tree(),
                tree,
                [("", ""), ("a", "a"), ("b/", "b/"), ("b/c", "c")],
            )
        tree._validate()

    def test_move_onto_self(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["b/", "b/a"])
        tree.add(["b", "b/a"])
        tree.commit("initial")

        self.assertRaises(errors.BzrMoveFailedError, tree.move, ["b/a"], "b")
        tree._validate()

    def test_move_onto_self_root(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a"])
        tree.add(["a"])
        tree.commit("initial")

        self.assertRaises(errors.BzrMoveFailedError, tree.move, ["a"], "a")
        tree._validate()

    def test_move_after(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/"])
        tree.add(["a", "b"])
        tree.commit("initial")
        os.rename("a", "b/a")

        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b/", "b/")]
        )
        # We don't need after=True as long as source is missing and target
        # exists.
        self.assertEqual([("a", "b/a")], tree.move(["a"], "b"))
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("b/", "b/"), ("b/a", "a")]
        )
        tree._validate()

    def test_move_after_with_after(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/"])
        tree.add(["a", "b"])
        tree.commit("initial")
        os.rename("a", "b/a")

        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("a", "a"), ("b/", "b/")]
        )
        # Passing after=True should work as well
        self.assertEqual([("a", "b/a")], tree.move(["a"], "b", after=True))
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("b/", "b/"), ("b/a", "a")]
        )
        tree._validate()

    def test_move_after_no_target(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/"])
        tree.add(["a", "b"])
        tree.commit("initial")

        # Passing after when the file hasn't been move raises an exception
        self.assertRaises(errors.BzrMoveFailedError, tree.move, ["a"], "b", after=True)

        self.assertTreeLayout(["", "a", "b/"], tree.basis_tree())
        tree._validate()

    def test_move_after_source_and_dest(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/a"])
        tree.add(["a", "b"])
        tree.commit("initial")

        # TODO: jam 20070225 I would usually use 'rb', but assertFileEqual
        #       uses 'r'.
        with open("a") as a_file:
            a_text = a_file.read()
        with open("b/a") as ba_file:
            ba_text = ba_file.read()

        self.assertTreeLayout(["", "a", "b/"], tree)
        self.assertRaises(
            errors.RenameFailedFilesExist, tree.move, ["a"], "b", after=False
        )
        self.assertTreeLayout(["", "a", "b/"], tree)
        self.assertFileEqual(a_text, "a")
        self.assertFileEqual(ba_text, "b/a")
        # But you can pass after=True
        self.assertEqual([("a", "b/a")], tree.move(["a"], "b", after=True))
        self.assertPathRelations(
            tree.basis_tree(), tree, [("", ""), ("b/", "b/"), ("b/a", "a")]
        )
        # But it shouldn't actually move anything
        self.assertFileEqual(a_text, "a")
        self.assertFileEqual(ba_text, "b/a")
        tree._validate()

    def test_move_directory(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "a/c/", "a/c/d", "e/"])
        tree.add(["a", "a/b", "a/c", "a/c/d", "e"])
        tree.commit("initial")

        self.assertEqual([("a", "e/a")], tree.move(["a"], "e"))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("e/", "e/"),
                ("e/a/", "a/"),
                ("e/a/b", "a/b"),
                ("e/a/c/", "a/c/"),
                ("e/a/c/d", "a/c/d"),
            ],
        )
        tree._validate()

    def test_move_directory_into_parent(self):
        if not self.workingtree_format.supports_versioned_directories:
            raise tests.TestNotApplicable("test requires versioned directories")
        tree = self.make_branch_and_tree(".")
        self.build_tree(["c/", "c/b/", "c/b/d/"])
        tree.add(["c", "c/b", "c/b/d"])
        tree.commit("initial")

        self.assertEqual([("c/b", "b")], tree.move(["c/b"], ""))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("b/", "c/b/"), ("c/", "c/"), ("b/d/", "c/b/d/")],
        )
        tree._validate()

    def test_move_directory_with_children_in_subdir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "a/c/", "d/"])
        tree.add(["a", "a/b", "a/c", "d"])
        tree.commit("initial")

        tree.rename_one("a/b", "a/c/b")
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("a/", "a/"),
                ("d/", "d/"),
                ("a/c/", "a/c/"),
                ("a/c/b", "a/b"),
            ],
        )
        self.assertEqual([("a", "d/a")], tree.move(["a"], "d"))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("d/", "d/"),
                ("d/a/", "a/"),
                ("d/a/c/", "a/c/"),
                ("d/a/c/b", "a/b"),
            ],
        )
        tree._validate()

    def test_move_directory_with_deleted_children(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "a/c", "a/d", "b/"])
        tree.add(["a", "b", "a/b", "a/c", "a/d"])
        tree.commit("initial")

        tree.remove(["a/b", "a/d"])

        self.assertEqual([("a", "b/a")], tree.move(["a"], "b"))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("b/", "b/"), ("b/a/", "a/"), ("b/a/c", "a/c")],
        )
        tree._validate()

    def test_move_directory_with_new_children(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/c", "b/"])
        tree.add(["a", "b", "a/c"])
        tree.commit("initial")

        self.build_tree(["a/b", "a/d"])
        tree.add(["a/b", "a/d"])

        self.assertEqual([("a", "b/a")], tree.move(["a"], "b"))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("b/", "b/"),
                ("b/a/", "a/"),
                ("b/a/b", None),
                ("b/a/c", "a/c"),
                ("b/a/d", None),
            ],
        )
        tree._validate()

    def test_move_directory_with_moved_children(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "a/c", "d", "e/"])
        tree.add(["a", "a/b", "a/c", "d", "e"])
        tree.commit("initial")

        self.assertEqual([("a/b", "b")], tree.move(["a/b"], ""))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("a/", "a/"),
                ("b", "a/b"),
                ("d", "d"),
                ("e/", "e/"),
                ("a/c", "a/c"),
            ],
        )
        self.assertEqual([("d", "a/d")], tree.move(["d"], "a"))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("a/", "a/"),
                ("b", "a/b"),
                ("e/", "e/"),
                ("a/c", "a/c"),
                ("a/d", "d"),
            ],
        )
        self.assertEqual([("a", "e/a")], tree.move(["a"], "e"))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("b", "a/b"),
                ("e/", "e/"),
                ("e/a/", "a/"),
                ("e/a/c", "a/c"),
                ("e/a/d", "d"),
            ],
        )
        tree._validate()

    def test_move_directory_with_renamed_child(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "a/c", "d/"])
        tree.add(["a", "a/b", "a/c", "d"])
        tree.commit("initial")

        tree.rename_one("a/b", "a/d")
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("a/", "a/"), ("d/", "d/"), ("a/c", "a/c"), ("a/d", "a/b")],
        )

        self.assertEqual([("a", "d/a")], tree.move(["a"], "d"))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("d/", "d/"),
                ("d/a/", "a/"),
                ("d/a/c", "a/c"),
                ("d/a/d", "a/b"),
            ],
        )
        tree._validate()

    def test_move_directory_with_swapped_children(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "a/c", "a/d", "e/"])
        tree.add(["a", "a/b", "a/c", "a/d", "e"])
        tree.commit("initial")

        tree.rename_one("a/b", "a/bb")
        tree.rename_one("a/d", "a/b")
        tree.rename_one("a/bb", "a/d")
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("a/", "a/"),
                ("e/", "e/"),
                ("a/b", "a/d"),
                ("a/c", "a/c"),
                ("a/d", "a/b"),
            ],
        )
        self.assertEqual([("a", "e/a")], tree.move(["a"], "e"))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [
                ("", ""),
                ("e/", "e/"),
                ("e/a/", "a/"),
                ("e/a/b", "a/d"),
                ("e/a/c", "a/c"),
                ("e/a/d", "a/b"),
            ],
        )
        tree._validate()

    def test_move_moved(self):
        """Moving a moved entry works as expected."""
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b", "c/"])
        tree.add(["a", "a/b", "c"])
        tree.commit("initial")

        self.assertEqual([("a/b", "c/b")], tree.move(["a/b"], "c"))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("a/", "a/"), ("c/", "c/"), ("c/b", "a/b")],
        )

        self.assertEqual([("c/b", "b")], tree.move(["c/b"], ""))
        self.assertPathRelations(
            tree.basis_tree(),
            tree,
            [("", ""), ("a/", "a/"), ("b", "a/b"), ("c/", "c/")],
        )
        tree._validate()

    def test_move_to_unversioned_non_ascii_dir(self):
        """Check error when moving to unversioned non-ascii directory."""
        self.requireFeature(features.UnicodeFilenameFeature)
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "\xa7/"])
        tree.add(["a"])
        if tree.has_versioned_directories():
            e = self.assertRaises(errors.BzrMoveFailedError, tree.move, ["a"], "\xa7")
            self.assertIsInstance(e.extra, errors.NotVersionedError)
            self.assertEqual(e.extra.path, "\xa7")
        else:
            tree.move(["a"], "\xa7")

    def test_move_unversioned_non_ascii(self):
        """Check error when moving an unversioned non-ascii file."""
        self.requireFeature(features.UnicodeFilenameFeature)
        tree = self.make_branch_and_tree(".")
        self.build_tree(["\xa7", "dir/"])
        tree.add("dir")
        e = self.assertRaises(errors.BzrMoveFailedError, tree.move, ["\xa7"], "dir")
        self.assertIsInstance(e.extra, errors.NotVersionedError)
        self.assertEqual(e.extra.path, "\xa7")



# Copyright (C) 2006-2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""External tests of 'brz ls'."""

from breezy import ignores, tests


class TestLS(tests.TestCaseWithTransport):
    def setUp(self):
        super().setUp()

        # Create a simple branch that can be used in testing
        ignores._set_user_ignores(["user-ignore"])

        self.wt = self.make_branch_and_tree(".")
        self.build_tree_contents(
            [
                (".bzrignore", b"*.pyo\n"),
                ("a", b"hello\n"),
            ]
        )

    def ls_equals(self, value, args=None, recursive=True, working_dir=None):
        command = "ls"
        if args is not None:
            command += " " + args
        if recursive:
            command += " -R"
        out, err = self.run_bzr(command, working_dir=working_dir)
        self.assertEqual("", err)
        self.assertEqualDiff(value, out)

    def test_ls_null_verbose(self):
        # Can't supply both
        self.run_bzr_error(
            ["Cannot set both --verbose and --null"], "ls --verbose --null"
        )

    def test_ls_basic(self):
        """Test the abilities of 'brz ls'."""
        self.ls_equals(".bzrignore\na\n")
        self.ls_equals(".bzrignore\na\n", "./")
        self.ls_equals("?        .bzrignore\n?        a\n", "--verbose")
        self.ls_equals(".bzrignore\na\n", "--unknown")
        self.ls_equals("", "--ignored")
        self.ls_equals("", "--versioned")
        self.ls_equals("", "-V")
        self.ls_equals(".bzrignore\na\n", "--unknown --ignored --versioned")
        self.ls_equals(".bzrignore\na\n", "--unknown --ignored -V")
        self.ls_equals("", "--ignored --versioned")
        self.ls_equals("", "--ignored -V")
        self.ls_equals(".bzrignore\0a\0", "--null")

    def test_ls_added(self):
        self.wt.add(["a"])
        self.ls_equals("?        .bzrignore\nV        a\n", "--verbose")
        self.wt.commit("add")

        self.build_tree(["subdir/"])
        self.ls_equals(
            "?        .bzrignore\nV        a\n?        subdir/\n", "--verbose"
        )
        self.build_tree(["subdir/b"])
        self.wt.add(["subdir/", "subdir/b", ".bzrignore"])
        self.ls_equals(
            "V        .bzrignore\nV        a\nV        subdir/\nV        subdir/b\n",
            "--verbose",
        )

    def test_show_ids(self):
        self.build_tree(["subdir/"])
        self.wt.add(["a", "subdir"], ids=[b"a-id", b"subdir-id"])
        self.ls_equals(
            ".bzrignore                                         \n"
            "a                                                  a-id\n"
            "subdir/                                            subdir-id\n",
            "--show-ids",
        )
        self.ls_equals(
            "?        .bzrignore\n"
            "V        a                                         a-id\n"
            "V        subdir/                                   subdir-id\n",
            "--show-ids --verbose",
        )
        self.ls_equals(
            ".bzrignore\0\0a\0a-id\0subdir\0subdir-id\0", "--show-ids --null"
        )

    def test_ls_no_recursive(self):
        self.build_tree(["subdir/", "subdir/b"])
        self.wt.add(["a", "subdir/", "subdir/b", ".bzrignore"])

        self.ls_equals(".bzrignore\na\nsubdir/\n", recursive=False)

        self.ls_equals(
            "V        .bzrignore\nV        a\nV        subdir/\n",
            "--verbose",
            recursive=False,
        )

        # Check what happens in a sub-directory
        self.ls_equals("b\n", working_dir="subdir")
        self.ls_equals("b\0", "--null", working_dir="subdir")
        self.ls_equals("subdir/b\n", "--from-root", working_dir="subdir")
        self.ls_equals("subdir/b\0", "--from-root --null", working_dir="subdir")
        self.ls_equals(
            "subdir/b\n", "--from-root", recursive=False, working_dir="subdir"
        )

    def test_ls_path(self):
        """If a path is specified, files are listed with that prefix."""
        self.build_tree(["subdir/", "subdir/b"])
        self.wt.add(["subdir", "subdir/b"])
        self.ls_equals("subdir/b\n", "subdir")
        self.ls_equals(
            "../.bzrignore\n../a\n../subdir/\n../subdir/b\n",
            "..",
            working_dir="subdir",
        )
        self.ls_equals(
            "../.bzrignore\0../a\0../subdir\0../subdir/b\0",
            ".. --null",
            working_dir="subdir",
        )
        self.ls_equals(
            "?        ../.bzrignore\n"
            "?        ../a\n"
            "V        ../subdir/\n"
            "V        ../subdir/b\n",
            ".. --verbose",
            working_dir="subdir",
        )
        self.run_bzr_error(
            ["cannot specify both --from-root and PATH"],
            "ls --from-root ..",
            working_dir="subdir",
        )

    def test_ls_revision(self):
        self.wt.add(["a"])
        self.wt.commit("add")

        self.build_tree(["subdir/"])

        # Check what happens when we supply a specific revision
        self.ls_equals("a\n", "--revision 1")
        self.ls_equals("V        a\n", "--verbose --revision 1")

        self.ls_equals("", "--revision 1", working_dir="subdir")

    def test_ls_branch(self):
        """If a branch is specified, files are listed from it."""
        self.build_tree(["subdir/", "subdir/b"])
        self.wt.add(["subdir", "subdir/b"])
        self.wt.commit("committing")
        branch = self.make_branch("branchdir")
        branch.pull(self.wt.branch)
        self.ls_equals("branchdir/subdir/\nbranchdir/subdir/b\n", "branchdir")
        self.ls_equals(
            "branchdir/subdir/\nbranchdir/subdir/b\n", "branchdir --revision 1"
        )

    def test_ls_ignored(self):
        # Now try to do ignored files.
        self.wt.add(["a", ".bzrignore"])

        self.build_tree(["blah.py", "blah.pyo", "user-ignore"])
        self.ls_equals(".bzrignore\na\nblah.py\nblah.pyo\nuser-ignore\n")
        self.ls_equals(
            "V        .bzrignore\n"
            "V        a\n"
            "?        blah.py\n"
            "I        blah.pyo\n"
            "I        user-ignore\n",
            "--verbose",
        )
        self.ls_equals("blah.pyo\nuser-ignore\n", "--ignored")
        self.ls_equals("blah.py\n", "--unknown")
        self.ls_equals(".bzrignore\na\n", "--versioned")
        self.ls_equals(".bzrignore\na\n", "-V")

    def test_kinds(self):
        self.build_tree(["subdir/"])
        self.ls_equals(".bzrignore\na\n", "--kind=file")
        self.ls_equals("subdir/\n", "--kind=directory")
        self.ls_equals("", "--kind=symlink")
        self.run_bzr_error(["invalid kind specified"], "ls --kind=pile")

    def test_ls_path_nonrecursive(self):
        self.ls_equals(
            "{}/.bzrignore\n{}/a\n".format(self.test_dir, self.test_dir),
            self.test_dir,
            recursive=False,
        )

    def test_ls_directory(self):
        """Test --directory option."""
        self.wt = self.make_branch_and_tree("dir")
        self.build_tree(["dir/sub/", "dir/sub/file"])
        self.wt.add(["sub", "sub/file"])
        self.wt.commit("commit")
        self.ls_equals("sub/\nsub/file\n", "--directory=dir")
        self.ls_equals("sub/file\n", "-d dir sub")



# Copyright (C) 2006-2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz remove-tree."""

import os

from breezy import shelf
from breezy.tests import TestCaseWithTransport


class TestRemoveTree(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree("branch1")
        self.build_tree(["branch1/foo"])
        self.tree.add("foo")
        self.tree.commit("1")
        self.assertPathExists("branch1/foo")

    # Success modes

    def test_remove_tree_original_branch(self):
        self.run_bzr("remove-tree", working_dir="branch1")
        self.assertPathDoesNotExist("branch1/foo")

    def test_remove_tree_original_branch_explicit(self):
        self.run_bzr("remove-tree branch1")
        self.assertPathDoesNotExist("branch1/foo")

    def test_remove_tree_multiple_branch_explicit(self):
        self.tree.controldir.sprout("branch2")
        self.run_bzr("remove-tree branch1 branch2")
        self.assertPathDoesNotExist("branch1/foo")
        self.assertPathDoesNotExist("branch2/foo")

    def test_remove_tree_sprouted_branch(self):
        self.tree.controldir.sprout("branch2")
        self.assertPathExists("branch2/foo")
        self.run_bzr("remove-tree", working_dir="branch2")
        self.assertPathDoesNotExist("branch2/foo")

    def test_remove_tree_sprouted_branch_explicit(self):
        self.tree.controldir.sprout("branch2")
        self.assertPathExists("branch2/foo")
        self.run_bzr("remove-tree branch2")
        self.assertPathDoesNotExist("branch2/foo")

    def test_remove_tree_checkout(self):
        self.tree.branch.create_checkout("branch2", lightweight=False)
        self.assertPathExists("branch2/foo")
        self.run_bzr("remove-tree", working_dir="branch2")
        self.assertPathDoesNotExist("branch2/foo")
        self.assertPathExists("branch1/foo")

    def test_remove_tree_checkout_explicit(self):
        self.tree.branch.create_checkout("branch2", lightweight=False)
        self.assertPathExists("branch2/foo")
        self.run_bzr("remove-tree branch2")
        self.assertPathDoesNotExist("branch2/foo")
        self.assertPathExists("branch1/foo")

    # Failure modes

    def test_remove_tree_lightweight_checkout(self):
        self.tree.branch.create_checkout("branch2", lightweight=True)
        self.assertPathExists("branch2/foo")
        self.run_bzr_error(
            ["You cannot remove the working tree from a lightweight checkout"],
            "remove-tree",
            retcode=3,
            working_dir="branch2",
        )
        self.assertPathExists("branch2/foo")
        self.assertPathExists("branch1/foo")

    def test_remove_tree_lightweight_checkout_explicit(self):
        self.tree.branch.create_checkout("branch2", lightweight=True)
        self.assertPathExists("branch2/foo")
        self.run_bzr_error(
            ["You cannot remove the working tree from a lightweight checkout"],
            "remove-tree branch2",
            retcode=3,
        )
        self.assertPathExists("branch2/foo")
        self.assertPathExists("branch1/foo")

    def test_remove_tree_empty_dir(self):
        os.mkdir("branch2")
        self.run_bzr_error(
            ["Not a branch"], "remove-tree", retcode=3, working_dir="branch2"
        )

    def test_remove_tree_repeatedly(self):
        self.run_bzr("remove-tree branch1")
        self.assertPathDoesNotExist("branch1/foo")
        self.run_bzr_error(
            ["No working tree to remove"], "remove-tree branch1", retcode=3
        )

    def test_remove_tree_remote_path(self):
        # TODO: I can't think of a way to implement this...
        pass

    def test_remove_tree_uncommitted_changes(self):
        self.build_tree(["branch1/bar"])
        self.tree.add("bar")
        self.run_bzr_error(
            ["Working tree .* has uncommitted changes"],
            "remove-tree branch1",
            retcode=3,
        )

    def test_remove_tree_uncommitted_changes_force(self):
        self.build_tree(["branch1/bar"])
        self.tree.add("bar")
        self.run_bzr("remove-tree branch1 --force")
        self.assertPathDoesNotExist("branch1/foo")
        self.assertPathExists("branch1/bar")

    def test_remove_tree_pending_merges(self):
        self.run_bzr(["branch", "branch1", "branch2"])
        self.build_tree(["branch1/bar"])
        self.tree.add("bar")
        self.tree.commit("2")
        self.assertPathExists("branch1/bar")
        self.run_bzr(["merge", "../branch1"], working_dir="branch2")
        self.assertPathExists("branch2/bar")
        self.run_bzr(["revert", "."], working_dir="branch2")
        self.assertPathDoesNotExist("branch2/bar")
        self.run_bzr_error(
            ["Working tree .* has uncommitted changes"],
            "remove-tree branch2",
            retcode=3,
        )

    def test_remove_tree_pending_merges_force(self):
        self.run_bzr(["branch", "branch1", "branch2"])
        self.build_tree(["branch1/bar"])
        self.tree.add("bar")
        self.tree.commit("2")
        self.assertPathExists("branch1/bar")
        self.run_bzr(["merge", "../branch1"], working_dir="branch2")
        self.assertPathExists("branch2/bar")
        self.run_bzr(["revert", "."], working_dir="branch2")
        self.assertPathDoesNotExist("branch2/bar")
        self.run_bzr("remove-tree branch2 --force")
        self.assertPathDoesNotExist("branch2/foo")
        self.assertPathDoesNotExist("branch2/bar")

    def test_remove_tree_shelved_changes(self):
        # https://bugs.launchpad.net/bzr/+bug/586639
        tree = self.make_branch_and_tree(".")
        creator = shelf.ShelfCreator(tree, tree.basis_tree(), [])
        self.addCleanup(creator.finalize)
        tree.get_shelf_manager().shelve_changes(creator, "Foo")
        self.run_bzr_error(
            ["Working tree .* has shelved changes"], "remove-tree", retcode=3
        )

    def test_remove_tree_shelved_changes_force(self):
        tree = self.make_branch_and_tree(".")
        creator = shelf.ShelfCreator(tree, tree.basis_tree(), [])
        self.addCleanup(creator.finalize)
        tree.get_shelf_manager().shelve_changes(creator, "Foo")
        self.run_bzr("remove-tree --force")
        self.run_bzr("checkout")
        # Ensure shelf is empty
        self.assertIs(None, tree.get_shelf_manager().last_shelf())



# Copyright (C) 2006-2012, 2015, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the urlutils wrapper."""

import os
import sys

from .. import osutils, urlutils
from ..errors import PathNotChild
from . import TestCase, TestCaseInTempDir, TestSkipped, features


class TestUrlToPath(TestCase):
    def test_basename(self):
        # breezy.urlutils.basename
        # Test breezy.urlutils.split()
        basename = urlutils.basename
        if sys.platform == "win32":
            self.assertRaises(urlutils.InvalidURL, basename, "file:///path/to/foo")
            self.assertEqual("foo", basename("file:///C|/foo"))
            self.assertEqual("foo", basename("file:///C:/foo"))
            self.assertEqual("", basename("file:///C:/"))
        else:
            self.assertEqual("foo", basename("file:///foo"))
            self.assertEqual("", basename("file:///"))

        self.assertEqual("foo", basename("http://host/path/to/foo"))
        self.assertEqual("foo", basename("http://host/path/to/foo/"))
        self.assertEqual(
            "", basename("http://host/path/to/foo/", exclude_trailing_slash=False)
        )
        self.assertEqual("path", basename("http://host/path"))
        self.assertEqual("", basename("http://host/"))
        self.assertEqual("", basename("http://host"))
        self.assertEqual("path", basename("http:///nohost/path"))

        self.assertEqual("path", basename("random+scheme://user:pass@ahost:port/path"))
        self.assertEqual("path", basename("random+scheme://user:pass@ahost:port/path/"))
        self.assertEqual("", basename("random+scheme://user:pass@ahost:port/"))

        # relative paths
        self.assertEqual("foo", basename("path/to/foo"))
        self.assertEqual("foo", basename("path/to/foo/"))
        self.assertEqual("", basename("path/to/foo/", exclude_trailing_slash=False))
        self.assertEqual("foo", basename("path/../foo"))
        self.assertEqual("foo", basename("../path/foo"))

    def test_normalize_url_files(self):
        # Test that local paths are properly normalized
        normalize_url = urlutils.normalize_url

        def norm_file(expected, path):
            url = normalize_url(path)
            self.assertStartsWith(url, "file:///")
            if sys.platform == "win32":
                url = url[len("file:///C:") :]
            else:
                url = url[len("file://") :]

            self.assertEndsWith(url, expected)

        norm_file("path/to/foo", "path/to/foo")
        norm_file("/path/to/foo", "/path/to/foo")
        norm_file("path/to/foo", "../path/to/foo")

        # Local paths are assumed to *not* be escaped at all
        try:
            "uni/\xb5".encode(osutils.get_user_encoding())
        except UnicodeError:
            # locale cannot handle unicode
            pass
        else:
            norm_file("uni/%C2%B5", "uni/\xb5")

        norm_file("uni/%25C2%25B5", "uni/%C2%B5")
        norm_file("uni/%20b", "uni/ b")
        # All the crazy characters get escaped in local paths => file:/// urls
        # The ' ' character must not be at the end, because on win32
        # it gets stripped off by ntpath.abspath
        norm_file("%27%20%3B/%3F%3A%40%26%3D%2B%24%2C%23", "' ;/?:@&=+$,#")

    def test_normalize_url_hybrid(self):
        # Anything with a scheme:// should be treated as a hybrid url
        # which changes what characters get escaped.
        normalize_url = urlutils.normalize_url

        eq = self.assertEqual
        eq("file:///foo/", normalize_url("file:///foo/"))
        eq("file:///foo/%20", normalize_url("file:///foo/ "))
        eq("file:///foo/%20", normalize_url("file:///foo/%20"))
        # Don't escape reserved characters
        eq(
            "file:///ab_c.d-e/%f:?g&h=i+j;k,L#M$",
            normalize_url("file:///ab_c.d-e/%f:?g&h=i+j;k,L#M$"),
        )
        eq(
            "http://ab_c.d-e/%f:?g&h=i+j;k,L#M$",
            normalize_url("http://ab_c.d-e/%f:?g&h=i+j;k,L#M$"),
        )

        # Escape unicode characters, but not already escaped chars
        eq("http://host/ab/%C2%B5/%C2%B5", normalize_url("http://host/ab/%C2%B5/\xb5"))

        # Unescape characters that don't need to be escaped
        eq(
            "http://host/~bob%2525-._",
            normalize_url("http://host/%7Ebob%2525%2D%2E%5F"),
        )
        eq(
            "http://host/~bob%2525-._",
            normalize_url("http://host/%7Ebob%2525%2D%2E%5F"),
        )

    def test_url_scheme_re(self):
        # Test paths that may be URLs
        def test_one(url, scheme_and_path):
            """Assert that _url_scheme_re correctly matches.

            :param scheme_and_path: The (scheme, path) that should be matched
                can be None, to indicate it should not match
            """
            m = urlutils._url_scheme_re.match(url)
            if scheme_and_path is None:
                self.assertEqual(None, m)
            else:
                self.assertEqual(scheme_and_path[0], m.group("scheme"))
                self.assertEqual(scheme_and_path[1], m.group("path"))

        # Local paths
        test_one("/path", None)
        test_one("C:/path", None)
        test_one("../path/to/foo", None)
        test_one("../path/to/fo\xe5", None)

        # Real URLS
        test_one("http://host/path/", ("http", "host/path/"))
        test_one("sftp://host/path/to/foo", ("sftp", "host/path/to/foo"))
        test_one("file:///usr/bin", ("file", "/usr/bin"))
        test_one("file:///C:/Windows", ("file", "/C:/Windows"))
        test_one("file:///C|/Windows", ("file", "/C|/Windows"))
        test_one("readonly+sftp://host/path/\xe5", ("readonly+sftp", "host/path/\xe5"))

        # Weird stuff
        # Can't have slashes or colons in the scheme
        test_one("/path/to/://foo", None)
        test_one("scheme:stuff://foo", ("scheme", "stuff://foo"))
        # Must have more than one character for scheme
        test_one("C://foo", None)
        test_one("ab://foo", ("ab", "foo"))

    def test_dirname(self):
        # Test breezy.urlutils.dirname()
        dirname = urlutils.dirname
        if sys.platform == "win32":
            self.assertRaises(urlutils.InvalidURL, dirname, "file:///path/to/foo")
            self.assertEqual("file:///C|/", dirname("file:///C|/foo"))
            self.assertEqual("file:///C|/", dirname("file:///C|/"))
        else:
            self.assertEqual("file:///", dirname("file:///foo"))
            self.assertEqual("file:///", dirname("file:///"))

        self.assertEqual("http://host/path/to", dirname("http://host/path/to/foo"))
        self.assertEqual("http://host/path/to", dirname("http://host/path/to/foo/"))
        self.assertEqual(
            "http://host/path/to/foo",
            dirname("http://host/path/to/foo/", exclude_trailing_slash=False),
        )
        self.assertEqual("http://host/", dirname("http://host/path"))
        self.assertEqual("http://host/", dirname("http://host/"))
        self.assertEqual("http://host", dirname("http://host"))
        self.assertEqual("http:///nohost", dirname("http:///nohost/path"))

        self.assertEqual(
            "random+scheme://user:pass@ahost:port/",
            dirname("random+scheme://user:pass@ahost:port/path"),
        )
        self.assertEqual(
            "random+scheme://user:pass@ahost:port/",
            dirname("random+scheme://user:pass@ahost:port/path/"),
        )
        self.assertEqual(
            "random+scheme://user:pass@ahost:port/",
            dirname("random+scheme://user:pass@ahost:port/"),
        )

        # relative paths
        self.assertEqual("path/to", dirname("path/to/foo"))
        self.assertEqual("path/to", dirname("path/to/foo/"))
        self.assertEqual(
            "path/to/foo", dirname("path/to/foo/", exclude_trailing_slash=False)
        )
        self.assertEqual("path/..", dirname("path/../foo"))
        self.assertEqual("../path", dirname("../path/foo"))

    def test_is_url(self):
        self.assertTrue(urlutils.is_url("http://foo/bar"))
        self.assertTrue(urlutils.is_url("bzr+ssh://foo/bar"))
        self.assertTrue(urlutils.is_url("lp:foo/bar"))
        self.assertTrue(urlutils.is_url("file:///foo/bar"))
        self.assertFalse(urlutils.is_url(""))
        self.assertFalse(urlutils.is_url("foo"))
        self.assertFalse(urlutils.is_url("foo/bar"))
        self.assertFalse(urlutils.is_url("/foo"))
        self.assertFalse(urlutils.is_url("/foo/bar"))
        self.assertFalse(urlutils.is_url("C:/"))
        self.assertFalse(urlutils.is_url("C:/foo"))
        self.assertFalse(urlutils.is_url("C:/foo/bar"))

    def test_join(self):
        def test(expected, *args):
            joined = urlutils.join(*args)
            self.assertEqual(expected, joined)

        # Test relative path joining
        test("foo", "foo")  # relative fragment with nothing is preserved.
        test("foo/bar", "foo", "bar")
        test("http://foo/bar", "http://foo", "bar")
        test("http://foo/bar", "http://foo", ".", "bar")
        test("http://foo/baz", "http://foo", "bar", "../baz")
        test("http://foo/bar/baz", "http://foo", "bar/baz")
        test("http://foo/baz", "http://foo", "bar/../baz")
        test("http://foo/baz", "http://foo/bar/", "../baz")
        test("lp:foo/bar", "lp:foo", "bar")
        test("lp:foo/bar/baz", "lp:foo", "bar/baz")

        # Absolute paths
        test("http://foo", "http://foo")  # abs url with nothing is preserved.
        test("http://bar", "http://foo", "http://bar")
        test("sftp://bzr/foo", "http://foo", "bar", "sftp://bzr/foo")
        test("file:///bar", "foo", "file:///bar")
        test("http://bar/", "http://foo", "http://bar/")
        test("http://bar/a", "http://foo", "http://bar/a")
        test("http://bar/a/", "http://foo", "http://bar/a/")
        test("lp:bar", "http://foo", "lp:bar")
        test("lp:bar", "lp:foo", "lp:bar")
        test("file:///stuff", "lp:foo", "file:///stuff")

        # From a base path
        test("file:///foo", "file:///", "foo")
        test("file:///bar/foo", "file:///bar/", "foo")
        test("http://host/foo", "http://host/", "foo")
        test("http://host/", "http://host", "")

        # Invalid joinings
        # Cannot go above root
        # Implicitly at root:
        self.assertRaises(
            urlutils.InvalidURLJoin, urlutils.join, "http://foo", "../baz"
        )
        self.assertRaises(urlutils.InvalidURLJoin, urlutils.join, "http://foo", "/..")
        # Joining from a path explicitly under the root.
        self.assertRaises(
            urlutils.InvalidURLJoin, urlutils.join, "http://foo/a", "../../b"
        )

    def test_joinpath(self):
        def test(expected, *args):
            joined = urlutils.joinpath(*args)
            self.assertEqual(expected, joined)

        # Test a single element
        test("foo", "foo")

        # Test relative path joining
        test("foo/bar", "foo", "bar")
        test("foo/bar", "foo", ".", "bar")
        test("foo/baz", "foo", "bar", "../baz")
        test("foo/bar/baz", "foo", "bar/baz")
        test("foo/baz", "foo", "bar/../baz")

        # Test joining to an absolute path
        test("/foo", "/foo")
        test("/foo", "/foo", ".")
        test("/foo/bar", "/foo", "bar")
        test("/", "/foo", "..")

        # Test joining with an absolute path
        test("/bar", "foo", "/bar")

        # Test joining to a path with a trailing slash
        test("foo/bar", "foo/", "bar")

        # Invalid joinings
        # Cannot go above root
        self.assertRaises(urlutils.InvalidURLJoin, urlutils.joinpath, "/", "../baz")
        self.assertRaises(urlutils.InvalidURLJoin, urlutils.joinpath, "/", "..")
        self.assertRaises(urlutils.InvalidURLJoin, urlutils.joinpath, "/", "/..")

    def test_join_segment_parameters_raw(self):
        join_segment_parameters_raw = urlutils.join_segment_parameters_raw
        self.assertEqual("/somedir/path", join_segment_parameters_raw("/somedir/path"))
        self.assertEqual(
            "/somedir/path,rawdata",
            join_segment_parameters_raw("/somedir/path", "rawdata"),
        )
        self.assertRaises(
            urlutils.InvalidURLJoin,
            join_segment_parameters_raw,
            "/somedir/path",
            "rawdata1,rawdata2,rawdata3",
        )
        self.assertEqual(
            "/somedir/path,bla,bar",
            join_segment_parameters_raw("/somedir/path", "bla", "bar"),
        )
        self.assertEqual(
            "/somedir,exist=some/path,bla,bar",
            join_segment_parameters_raw("/somedir,exist=some/path", "bla", "bar"),
        )
        self.assertRaises(TypeError, join_segment_parameters_raw, "/somepath", 42)

    def test_join_segment_parameters(self):
        join_segment_parameters = urlutils.join_segment_parameters
        self.assertEqual("/somedir/path", join_segment_parameters("/somedir/path", {}))
        self.assertEqual(
            "/somedir/path,key1=val1",
            join_segment_parameters("/somedir/path", {"key1": "val1"}),
        )
        self.assertRaises(
            urlutils.InvalidURLJoin,
            join_segment_parameters,
            "/somedir/path",
            {"branch": "brr,brr,brr"},
        )
        self.assertRaises(
            urlutils.InvalidURLJoin,
            join_segment_parameters,
            "/somedir/path",
            {"key1=val1": "val2"},
        )
        self.assertEqual(
            "/somedir/path,key1=val1,key2=val2",
            join_segment_parameters("/somedir/path", {"key1": "val1", "key2": "val2"}),
        )
        self.assertEqual(
            "/somedir/path,key1=val1,key2=val2",
            join_segment_parameters("/somedir/path,key1=val1", {"key2": "val2"}),
        )
        self.assertEqual(
            "/somedir/path,key1=val2",
            join_segment_parameters("/somedir/path,key1=val1", {"key1": "val2"}),
        )
        self.assertEqual(
            "/somedir,exist=some/path,key1=val1",
            join_segment_parameters("/somedir,exist=some/path", {"key1": "val1"}),
        )
        self.assertEqual(
            "/,key1=val1,key2=val2",
            join_segment_parameters("/,key1=val1", {"key2": "val2"}),
        )
        self.assertRaises(
            TypeError, join_segment_parameters, "/,key1=val1", {"foo": 42}
        )

    def test_function_type(self):
        if sys.platform == "win32":
            self.assertEqual(
                urlutils._win32_local_path_to_url, urlutils.local_path_to_url
            )
            self.assertEqual(
                urlutils._win32_local_path_from_url, urlutils.local_path_from_url
            )
        else:
            self.assertEqual(
                urlutils._posix_local_path_to_url, urlutils.local_path_to_url
            )
            self.assertEqual(
                urlutils._posix_local_path_from_url, urlutils.local_path_from_url
            )

    def test_posix_local_path_to_url(self):
        to_url = urlutils._posix_local_path_to_url
        self.assertEqual("file:///path/to/foo", to_url("/path/to/foo"))

        self.assertEqual("file:///path/to/foo%2Cbar", to_url("/path/to/foo,bar"))

        try:
            result = to_url("/path/to/r\xe4ksm\xf6rg\xe5s")
        except UnicodeError:
            raise TestSkipped("local encoding cannot handle unicode")

        self.assertEqual("file:///path/to/r%C3%A4ksm%C3%B6rg%C3%A5s", result)
        self.assertTrue(isinstance(result, str))

    def test_posix_local_path_from_url(self):
        from_url = urlutils._posix_local_path_from_url
        self.assertEqual("/path/to/foo", from_url("file:///path/to/foo"))
        self.assertEqual("/path/to/foo", from_url("file:///path/to/foo,branch=foo"))
        self.assertEqual(
            "/path/to/r\xe4ksm\xf6rg\xe5s",
            from_url("file:///path/to/r%C3%A4ksm%C3%B6rg%C3%A5s"),
        )
        self.assertEqual(
            "/path/to/r\xe4ksm\xf6rg\xe5s",
            from_url("file:///path/to/r%c3%a4ksm%c3%b6rg%c3%a5s"),
        )
        self.assertEqual(
            "/path/to/r\xe4ksm\xf6rg\xe5s",
            from_url("file://localhost/path/to/r%c3%a4ksm%c3%b6rg%c3%a5s"),
        )

        self.assertRaises(urlutils.InvalidURL, from_url, "/path/to/foo")
        self.assertRaises(
            urlutils.InvalidURL,
            from_url,
            "file://remotehost/path/to/r%c3%a4ksm%c3%b6rg%c3%a5s",
        )

    def test_win32_local_path_to_url(self):
        to_url = urlutils._win32_local_path_to_url
        self.assertEqual("file:///C:/path/to/foo", to_url("C:/path/to/foo"))
        # BOGUS: on win32, ntpath.abspath will strip trailing
        #       whitespace, so this will always fail
        #       Though under linux, it fakes abspath support
        #       and thus will succeed
        # self.assertEqual('file:///C:/path/to/foo%20',
        #     to_url('C:/path/to/foo '))
        self.assertEqual("file:///C:/path/to/f%20oo", to_url("C:/path/to/f oo"))

        self.assertEqual("file:///", to_url("/"))

        self.assertEqual("file:///C:/path/to/foo%2Cbar", to_url("C:/path/to/foo,bar"))
        try:
            result = to_url("d:/path/to/r\xe4ksm\xf6rg\xe5s")
        except UnicodeError:
            raise TestSkipped("local encoding cannot handle unicode")

        self.assertEqual("file:///D:/path/to/r%C3%A4ksm%C3%B6rg%C3%A5s", result)
        self.assertIsInstance(result, str)

    def test_win32_unc_path_to_url(self):
        self.requireFeature(features.win32_feature)
        to_url = urlutils._win32_local_path_to_url
        self.assertEqual("file://HOST/path", to_url(r"\\HOST\path"))
        self.assertEqual("file://HOST/path", to_url("//HOST/path"))

        try:
            result = to_url("//HOST/path/to/r\xe4ksm\xf6rg\xe5s")
        except UnicodeError:
            raise TestSkipped("local encoding cannot handle unicode")

        self.assertEqual("file://HOST/path/to/r%C3%A4ksm%C3%B6rg%C3%A5s", result)
        self.assertFalse(isinstance(result, str))

    def test_win32_local_path_from_url(self):
        from_url = urlutils._win32_local_path_from_url
        self.assertEqual("C:/path/to/foo", from_url("file:///C|/path/to/foo"))
        self.assertEqual(
            "D:/path/to/r\xe4ksm\xf6rg\xe5s",
            from_url("file:///d|/path/to/r%C3%A4ksm%C3%B6rg%C3%A5s"),
        )
        self.assertEqual(
            "D:/path/to/r\xe4ksm\xf6rg\xe5s",
            from_url("file:///d:/path/to/r%c3%a4ksm%c3%b6rg%c3%a5s"),
        )
        self.assertEqual("/", from_url("file:///"))
        self.assertEqual(
            "C:/path/to/foo", from_url("file:///C|/path/to/foo,branch=foo")
        )

        self.assertRaises(urlutils.InvalidURL, from_url, "file:///C:")
        self.assertRaises(urlutils.InvalidURL, from_url, "file:///c")
        self.assertRaises(urlutils.InvalidURL, from_url, "/path/to/foo")
        # Not a valid _win32 url, no drive letter
        self.assertRaises(urlutils.InvalidURL, from_url, "file:///path/to/foo")

    def test_win32_unc_path_from_url(self):
        from_url = urlutils._win32_local_path_from_url
        self.assertEqual("//HOST/path", from_url("file://HOST/path"))
        self.assertEqual("//HOST/path", from_url("file://HOST/path,branch=foo"))
        # despite IE allows 2, 4, 5 and 6 slashes in URL to another machine
        # we want to use only 2 slashes
        # Firefox understand only 5 slashes in URL, but it's ugly
        self.assertRaises(urlutils.InvalidURL, from_url, "file:////HOST/path")
        self.assertRaises(urlutils.InvalidURL, from_url, "file://///HOST/path")
        self.assertRaises(urlutils.InvalidURL, from_url, "file://////HOST/path")
        # check for file://C:/ instead of file:///C:/
        self.assertRaises(urlutils.InvalidURL, from_url, "file://C:/path")

    def test_win32_extract_drive_letter(self):
        extract = urlutils._win32_extract_drive_letter
        self.assertEqual(("file:///C:", "/foo"), extract("file://", "/C:/foo"))
        self.assertEqual(("file:///d|", "/path"), extract("file://", "/d|/path"))
        self.assertRaises(urlutils.InvalidURL, extract, "file://", "/path")
        # Root drives without slash treated as invalid, see bug #841322
        self.assertEqual(("file:///C:", "/"), extract("file://", "/C:/"))
        self.assertRaises(urlutils.InvalidURL, extract, "file://", "/C:")
        # Invalid without drive separator or following forward slash
        self.assertRaises(urlutils.InvalidURL, extract, "file://", "/C")
        self.assertRaises(urlutils.InvalidURL, extract, "file://", "/C:ool")

    def test_split(self):
        # Test breezy.urlutils.split()
        split = urlutils.split
        if sys.platform == "win32":
            self.assertRaises(urlutils.InvalidURL, split, "file:///path/to/foo")
            self.assertEqual(("file:///C|/", "foo"), split("file:///C|/foo"))
            self.assertEqual(("file:///C:/", ""), split("file:///C:/"))
        else:
            self.assertEqual(("file:///", "foo"), split("file:///foo"))
            self.assertEqual(("file:///", ""), split("file:///"))

        self.assertEqual(
            ("http://host/path/to", "foo"), split("http://host/path/to/foo")
        )
        self.assertEqual(
            ("http://host/path/to", "foo"), split("http://host/path/to/foo/")
        )
        self.assertEqual(
            ("http://host/path/to/foo", ""),
            split("http://host/path/to/foo/", exclude_trailing_slash=False),
        )
        self.assertEqual(("http://host/", "path"), split("http://host/path"))
        self.assertEqual(("http://host/", ""), split("http://host/"))
        self.assertEqual(("http://host", ""), split("http://host"))
        self.assertEqual(("http:///nohost", "path"), split("http:///nohost/path"))

        self.assertEqual(
            ("random+scheme://user:pass@ahost:port/", "path"),
            split("random+scheme://user:pass@ahost:port/path"),
        )
        self.assertEqual(
            ("random+scheme://user:pass@ahost:port/", "path"),
            split("random+scheme://user:pass@ahost:port/path/"),
        )
        self.assertEqual(
            ("random+scheme://user:pass@ahost:port/", ""),
            split("random+scheme://user:pass@ahost:port/"),
        )

        # relative paths
        self.assertEqual(("path/to", "foo"), split("path/to/foo"))
        self.assertEqual(("path/to", "foo"), split("path/to/foo/"))
        self.assertEqual(
            ("path/to/foo", ""), split("path/to/foo/", exclude_trailing_slash=False)
        )
        self.assertEqual(("path/..", "foo"), split("path/../foo"))
        self.assertEqual(("../path", "foo"), split("../path/foo"))

    def test_strip_segment_parameters(self):
        strip_segment_parameters = urlutils.strip_segment_parameters
        # Check relative references with absolute paths
        self.assertEqual("/some/path", strip_segment_parameters("/some/path"))
        self.assertEqual("/some/path", strip_segment_parameters("/some/path,tip"))
        self.assertEqual(
            "/some,dir/path", strip_segment_parameters("/some,dir/path,tip")
        )
        self.assertEqual(
            "/somedir/path", strip_segment_parameters("/somedir/path,heads%2Ftip")
        )
        self.assertEqual(
            "/somedir/path", strip_segment_parameters("/somedir/path,heads%2Ftip,bar")
        )
        # Check relative references with relative paths
        self.assertEqual("", strip_segment_parameters(",key1=val1"))
        self.assertEqual("foo/", strip_segment_parameters("foo/,key1=val1"))
        self.assertEqual("foo", strip_segment_parameters("foo,key1=val1"))
        self.assertEqual(
            "foo/base,la=bla/other/elements",
            strip_segment_parameters("foo/base,la=bla/other/elements"),
        )
        self.assertEqual(
            "foo/base,la=bla/other/elements",
            strip_segment_parameters("foo/base,la=bla/other/elements,a=b"),
        )
        # TODO: Check full URLs as well as relative references

    def test_split_segment_parameters_raw(self):
        split_segment_parameters_raw = urlutils.split_segment_parameters_raw
        # Check relative references with absolute paths
        self.assertEqual(("/some/path", []), split_segment_parameters_raw("/some/path"))
        self.assertEqual(
            ("/some/path", ["tip"]), split_segment_parameters_raw("/some/path,tip")
        )
        self.assertEqual(
            ("/some,dir/path", ["tip"]),
            split_segment_parameters_raw("/some,dir/path,tip"),
        )
        self.assertEqual(
            ("/somedir/path", ["heads%2Ftip"]),
            split_segment_parameters_raw("/somedir/path,heads%2Ftip"),
        )
        self.assertEqual(
            ("/somedir/path", ["heads%2Ftip", "bar"]),
            split_segment_parameters_raw("/somedir/path,heads%2Ftip,bar"),
        )
        # Check relative references with relative paths
        self.assertEqual(
            ("", ["key1=val1"]), split_segment_parameters_raw(",key1=val1")
        )
        self.assertEqual(
            ("foo/", ["key1=val1"]), split_segment_parameters_raw("foo/,key1=val1")
        )
        self.assertEqual(
            ("foo", ["key1=val1"]), split_segment_parameters_raw("foo,key1=val1")
        )
        self.assertEqual(
            ("foo/base,la=bla/other/elements", []),
            split_segment_parameters_raw("foo/base,la=bla/other/elements"),
        )
        self.assertEqual(
            ("foo/base,la=bla/other/elements", ["a=b"]),
            split_segment_parameters_raw("foo/base,la=bla/other/elements,a=b"),
        )
        # TODO: Check full URLs as well as relative references

    def test_split_segment_parameters(self):
        split_segment_parameters = urlutils.split_segment_parameters
        # Check relative references with absolute paths
        self.assertEqual(("/some/path", {}), split_segment_parameters("/some/path"))
        self.assertEqual(
            ("/some/path", {"branch": "tip"}),
            split_segment_parameters("/some/path,branch=tip"),
        )
        self.assertEqual(
            ("/some,dir/path", {"branch": "tip"}),
            split_segment_parameters("/some,dir/path,branch=tip"),
        )
        self.assertEqual(
            ("/somedir/path", {"ref": "heads%2Ftip"}),
            split_segment_parameters("/somedir/path,ref=heads%2Ftip"),
        )
        self.assertEqual(
            ("/somedir/path", {"ref": "heads%2Ftip", "key1": "val1"}),
            split_segment_parameters("/somedir/path,ref=heads%2Ftip,key1=val1"),
        )
        self.assertEqual(
            ("/somedir/path", {"ref": "heads%2F=tip"}),
            split_segment_parameters("/somedir/path,ref=heads%2F=tip"),
        )
        # Check relative references with relative paths
        self.assertEqual(("", {"key1": "val1"}), split_segment_parameters(",key1=val1"))
        self.assertEqual(
            ("foo/", {"key1": "val1"}), split_segment_parameters("foo/,key1=val1")
        )
        self.assertEqual(
            ("foo/base,key1=val1/other/elements", {}),
            split_segment_parameters("foo/base,key1=val1/other/elements"),
        )
        self.assertEqual(
            ("foo/base,key1=val1/other/elements", {"key2": "val2"}),
            split_segment_parameters("foo/base,key1=val1/other/elements,key2=val2"),
        )
        self.assertRaises(
            urlutils.InvalidURL, split_segment_parameters, "foo/base,key1"
        )
        # TODO: Check full URLs as well as relative references

    def test_win32_strip_local_trailing_slash(self):
        strip = urlutils._win32_strip_local_trailing_slash
        self.assertEqual("file://", strip("file://"))
        self.assertEqual("file:///", strip("file:///"))
        self.assertEqual("file:///C", strip("file:///C"))
        self.assertEqual("file:///C:", strip("file:///C:"))
        self.assertEqual("file:///d|", strip("file:///d|"))
        self.assertEqual("file:///C:/", strip("file:///C:/"))
        self.assertEqual("file:///C:/a", strip("file:///C:/a/"))

    def test_strip_trailing_slash(self):
        sts = urlutils.strip_trailing_slash
        if sys.platform == "win32":
            self.assertEqual("file:///C|/", sts("file:///C|/"))
            self.assertEqual("file:///C:/foo", sts("file:///C:/foo"))
            self.assertEqual("file:///C|/foo", sts("file:///C|/foo/"))
        else:
            self.assertEqual("file:///", sts("file:///"))
            self.assertEqual("file:///foo", sts("file:///foo"))
            self.assertEqual("file:///foo", sts("file:///foo/"))

        self.assertEqual("http://host/", sts("http://host/"))
        self.assertEqual("http://host/foo", sts("http://host/foo"))
        self.assertEqual("http://host/foo", sts("http://host/foo/"))

        # No need to fail just because the slash is missing
        self.assertEqual("http://host", sts("http://host"))
        # TODO: jam 20060502 Should this raise InvalidURL?
        self.assertEqual("file://", sts("file://"))

        self.assertEqual(
            "random+scheme://user:pass@ahost:port/path",
            sts("random+scheme://user:pass@ahost:port/path"),
        )
        self.assertEqual(
            "random+scheme://user:pass@ahost:port/path",
            sts("random+scheme://user:pass@ahost:port/path/"),
        )
        self.assertEqual(
            "random+scheme://user:pass@ahost:port/",
            sts("random+scheme://user:pass@ahost:port/"),
        )

        # Make sure relative paths work too
        self.assertEqual("path/to/foo", sts("path/to/foo"))
        self.assertEqual("path/to/foo", sts("path/to/foo/"))
        self.assertEqual("../to/foo", sts("../to/foo/"))
        self.assertEqual("path/../foo", sts("path/../foo/"))

    def test_unescape_for_display_utf8(self):
        # Test that URLs are converted to nice unicode strings for display
        def test(expected, url, encoding="utf-8"):
            disp_url = urlutils.unescape_for_display(url, encoding=encoding)
            self.assertIsInstance(disp_url, str)
            self.assertEqual(expected, disp_url)

        test("http://foo", "http://foo")
        if sys.platform == "win32":
            test("C:/foo/path", "file:///C|/foo/path")
            test("C:/foo/path", "file:///C:/foo/path")
        else:
            test("/foo/path", "file:///foo/path")

        test("http://foo/%2Fbaz", "http://foo/%2Fbaz")
        test("http://host/r\xe4ksm\xf6rg\xe5s", "http://host/r%C3%A4ksm%C3%B6rg%C3%A5s")

        # Make sure special escaped characters stay escaped
        test(
            "http://host/%3B%2F%3F%3A%40%26%3D%2B%24%2C%23",
            "http://host/%3B%2F%3F%3A%40%26%3D%2B%24%2C%23",
        )

        # Can we handle sections that don't have utf-8 encoding?
        test(
            "http://host/%EE%EE%EE/r\xe4ksm\xf6rg\xe5s",
            "http://host/%EE%EE%EE/r%C3%A4ksm%C3%B6rg%C3%A5s",
        )

        # Test encoding into output that can handle some characters
        test(
            "http://host/%EE%EE%EE/r\xe4ksm\xf6rg\xe5s",
            "http://host/%EE%EE%EE/r%C3%A4ksm%C3%B6rg%C3%A5s",
            encoding="iso-8859-1",
        )

        # This one can be encoded into utf8
        test(
            "http://host/\u062c\u0648\u062c\u0648",
            "http://host/%d8%ac%d9%88%d8%ac%d9%88",
            encoding="utf-8",
        )

        # This can't be put into 8859-1 and so stays as escapes
        test(
            "http://host/%d8%ac%d9%88%d8%ac%d9%88",
            "http://host/%d8%ac%d9%88%d8%ac%d9%88",
            encoding="iso-8859-1",
        )

    def test_escape(self):
        self.assertEqual("%25", urlutils.escape("%"))
        self.assertEqual("%C3%A5", urlutils.escape("\xe5"))
        self.assertIsInstance(urlutils.escape("\xe5"), str)

    def test_escape_tildes(self):
        self.assertEqual("~foo", urlutils.escape("~foo"))

    def test_unescape(self):
        self.assertEqual("%", urlutils.unescape("%25"))
        self.assertEqual("\xe5", urlutils.unescape("%C3%A5"))

        self.assertEqual("\xe5", urlutils.unescape("%C3%A5"))

    def test_escape_unescape(self):
        self.assertEqual("\xe5", urlutils.unescape(urlutils.escape("\xe5")))
        self.assertEqual("%", urlutils.unescape(urlutils.escape("%")))

    def test_relative_url(self):
        def test(expected, base, other):
            result = urlutils.relative_url(base, other)
            self.assertEqual(expected, result)

        test("a", "http://host/", "http://host/a")
        test(
            "http://entirely/different",
            "sftp://host/branch",
            "http://entirely/different",
        )
        test(
            "../person/feature",
            "http://host/branch/mainline",
            "http://host/branch/person/feature",
        )
        test("..", "http://host/branch", "http://host/")
        test("http://host2/branch", "http://host1/branch", "http://host2/branch")
        test(".", "http://host1/branch", "http://host1/branch")
        test(
            "../../../branch/2b",
            "file:///home/jelmer/foo/bar/2b",
            "file:///home/jelmer/branch/2b",
        )
        test(
            "../../branch/2b",
            "sftp://host/home/jelmer/bar/2b",
            "sftp://host/home/jelmer/branch/2b",
        )
        test(
            "../../branch/feature/%2b",
            "http://host/home/jelmer/bar/%2b",
            "http://host/home/jelmer/branch/feature/%2b",
        )
        test(
            "../../branch/feature/2b",
            "http://host/home/jelmer/bar/2b/",
            "http://host/home/jelmer/branch/feature/2b",
        )
        # relative_url should preserve a trailing slash
        test(
            "../../branch/feature/2b/",
            "http://host/home/jelmer/bar/2b/",
            "http://host/home/jelmer/branch/feature/2b/",
        )
        test(
            "../../branch/feature/2b/",
            "http://host/home/jelmer/bar/2b",
            "http://host/home/jelmer/branch/feature/2b/",
        )

        # TODO: treat http://host as http://host/
        #       relative_url is typically called from a branch.base or
        #       transport.base which always ends with a /
        # test('a', 'http://host', 'http://host/a')
        test("http://host/a", "http://host", "http://host/a")
        # test('.', 'http://host', 'http://host/')
        test("http://host/", "http://host", "http://host/")
        # test('.', 'http://host/', 'http://host')
        test("http://host", "http://host/", "http://host")

        # On Windows file:///C:/path/to and file:///D:/other/path
        # should not use relative url over the non-existent '/' directory.
        if sys.platform == "win32":
            # on the same drive
            test("../../other/path", "file:///C:/path/to", "file:///C:/other/path")
            # ~next two tests is failed, i.e. urlutils.relative_url expects
            # ~to see normalized file URLs?
            # ~test('../../other/path',
            # ~    'file:///C:/path/to', 'file:///c:/other/path')
            # ~test('../../other/path',
            # ~    'file:///C:/path/to', 'file:///C|/other/path')

            # check UNC paths too
            test(
                "../../other/path",
                "file://HOST/base/path/to",
                "file://HOST/base/other/path",
            )
            # on different drives
            test("file:///D:/other/path", "file:///C:/path/to", "file:///D:/other/path")
            # TODO: strictly saying in UNC path //HOST/base is full analog
            # of drive letter for hard disk, and this situation is also
            # should be exception from rules. [bialix 20071221]


class TestCwdToURL(TestCaseInTempDir):
    """Test that local_path_to_url works based on the cwd."""

    def test_dot(self):
        # This test will fail if getcwd is not ascii
        os.mkdir("mytest")
        os.chdir("mytest")

        url = urlutils.local_path_to_url(".")
        self.assertEndsWith(url, "/mytest")

    def test_non_ascii(self):
        try:
            os.mkdir("dod\xe9")
        except UnicodeError:
            raise TestSkipped("cannot create unicode directory")

        os.chdir("dod\xe9")

        # On Mac OSX this directory is actually:
        #   u'/dode\u0301' => '/dode\xcc\x81
        # but we should normalize it back to
        #   u'/dod\xe9' => '/dod\xc3\xa9'
        url = urlutils.local_path_to_url(".")
        self.assertEndsWith(url, "/dod%C3%A9")


class TestDeriveToLocation(TestCase):
    """Test that the mapping of FROM_LOCATION to TO_LOCATION works."""

    def test_to_locations_derived_from_paths(self):
        derive = urlutils.derive_to_location
        self.assertEqual("bar", derive("bar"))
        self.assertEqual("bar", derive("../bar"))
        self.assertEqual("bar", derive("/foo/bar"))
        self.assertEqual("bar", derive("c:/foo/bar"))
        self.assertEqual("bar", derive("c:bar"))

    def test_to_locations_derived_from_urls(self):
        derive = urlutils.derive_to_location
        self.assertEqual("bar", derive("http://foo/bar"))
        self.assertEqual("bar", derive("bzr+ssh://foo/bar"))
        self.assertEqual("foo-bar", derive("lp:foo-bar"))


class TestRebaseURL(TestCase):
    """Test the behavior of rebase_url."""

    def test_non_relative(self):
        result = urlutils.rebase_url("file://foo", "file://foo", "file://foo/bar")
        self.assertEqual("file://foo", result)
        result = urlutils.rebase_url("/foo", "file://foo", "file://foo/bar")
        self.assertEqual("/foo", result)

    def test_different_ports(self):
        e = self.assertRaises(
            urlutils.InvalidRebaseURLs,
            urlutils.rebase_url,
            "foo",
            "http://bar:80",
            "http://bar:81",
        )
        self.assertEqual(
            str(e),
            "URLs differ by more than path: 'http://bar:80' and 'http://bar:81'",
        )

    def test_different_hosts(self):
        e = self.assertRaises(
            urlutils.InvalidRebaseURLs,
            urlutils.rebase_url,
            "foo",
            "http://bar",
            "http://baz",
        )
        self.assertEqual(
            str(e), "URLs differ by more than path: 'http://bar' and 'http://baz'"
        )

    def test_different_protocol(self):
        e = self.assertRaises(
            urlutils.InvalidRebaseURLs,
            urlutils.rebase_url,
            "foo",
            "http://bar",
            "ftp://bar",
        )
        self.assertEqual(
            str(e), "URLs differ by more than path: 'http://bar' and 'ftp://bar'"
        )

    def test_rebase_success(self):
        self.assertEqual(
            "../bar", urlutils.rebase_url("bar", "http://baz/", "http://baz/qux")
        )
        self.assertEqual(
            "qux/bar", urlutils.rebase_url("bar", "http://baz/qux", "http://baz/")
        )
        self.assertEqual(
            ".", urlutils.rebase_url("foo", "http://bar/", "http://bar/foo/")
        )
        self.assertEqual(
            "qux/bar",
            urlutils.rebase_url("../bar", "http://baz/qux/foo", "http://baz/"),
        )

    def test_determine_relative_path(self):
        self.assertEqual(
            "../../baz/bar", urlutils.determine_relative_path("/qux/quxx", "/baz/bar")
        )
        self.assertEqual("..", urlutils.determine_relative_path("/bar/baz", "/bar"))
        self.assertEqual("baz", urlutils.determine_relative_path("/bar", "/bar/baz"))
        self.assertEqual(".", urlutils.determine_relative_path("/bar", "/bar"))


class TestParseURL(TestCase):
    def test_parse_simple(self):
        parsed = urlutils.parse_url("http://example.com:80/one")
        self.assertEqual(("http", None, None, "example.com", 80, "/one"), parsed)

    def test_ipv6(self):
        parsed = urlutils.parse_url("http://[1:2:3::40]/one")
        self.assertEqual(("http", None, None, "1:2:3::40", None, "/one"), parsed)

    def test_ipv6_port(self):
        parsed = urlutils.parse_url("http://[1:2:3::40]:80/one")
        self.assertEqual(("http", None, None, "1:2:3::40", 80, "/one"), parsed)


class TestURL(TestCase):
    def test_parse_simple(self):
        parsed = urlutils.URL.from_string("http://example.com:80/one")
        self.assertEqual("http", parsed.scheme)
        self.assertIs(None, parsed.user)
        self.assertIs(None, parsed.password)
        self.assertEqual("example.com", parsed.host)
        self.assertEqual(80, parsed.port)
        self.assertEqual("/one", parsed.path)

    def test_ipv6(self):
        parsed = urlutils.URL.from_string("http://[1:2:3::40]/one")
        self.assertEqual("http", parsed.scheme)
        self.assertIs(None, parsed.port)
        self.assertIs(None, parsed.user)
        self.assertIs(None, parsed.password)
        self.assertEqual("1:2:3::40", parsed.host)
        self.assertEqual("/one", parsed.path)

    def test_ipv6_port(self):
        parsed = urlutils.URL.from_string("http://[1:2:3::40]:80/one")
        self.assertEqual("http", parsed.scheme)
        self.assertEqual("1:2:3::40", parsed.host)
        self.assertIs(None, parsed.user)
        self.assertIs(None, parsed.password)
        self.assertEqual(80, parsed.port)
        self.assertEqual("/one", parsed.path)

    def test_quoted(self):
        parsed = urlutils.URL.from_string(
            "http://ro%62ey:h%40t@ex%41mple.com:2222/path"
        )
        self.assertEqual(parsed.quoted_host, "ex%41mple.com")
        self.assertEqual(parsed.host, "exAmple.com")
        self.assertEqual(parsed.port, 2222)
        self.assertEqual(parsed.quoted_user, "ro%62ey")
        self.assertEqual(parsed.user, "robey")
        self.assertEqual(parsed.quoted_password, "h%40t")
        self.assertEqual(parsed.password, "h@t")
        self.assertEqual(parsed.path, "/path")

    def test_eq(self):
        parsed1 = urlutils.URL.from_string("http://[1:2:3::40]:80/one")
        parsed2 = urlutils.URL.from_string("http://[1:2:3::40]:80/one")
        self.assertEqual(parsed1, parsed2)
        self.assertEqual(parsed1, parsed1)
        parsed2.path = "/two"
        self.assertNotEqual(parsed1, parsed2)

    def test_repr(self):
        parsed = urlutils.URL.from_string("http://[1:2:3::40]:80/one")
        self.assertEqual(
            "<URL('http', None, None, '1:2:3::40', 80, '/one')>", repr(parsed)
        )

    def test_str(self):
        parsed = urlutils.URL.from_string("http://[1:2:3::40]:80/one")
        self.assertEqual("http://[1:2:3::40]:80/one", str(parsed))

    def test__combine_paths(self):
        combine = urlutils.URL._combine_paths
        self.assertEqual(
            "/home/sarah/project/foo", combine("/home/sarah", "project/foo")
        )
        self.assertEqual("/etc", combine("/home/sarah", "../../etc"))
        self.assertEqual("/etc", combine("/home/sarah", "../../../etc"))
        self.assertEqual("/etc", combine("/home/sarah", "/etc"))

    def test_clone(self):
        url = urlutils.URL.from_string("http://[1:2:3::40]:80/one")
        url1 = url.clone("two")
        self.assertEqual("/one/two", url1.path)
        url2 = url.clone("/two")
        self.assertEqual("/two", url2.path)
        url3 = url.clone()
        self.assertIsNot(url, url3)
        self.assertEqual(url, url3)

    def test_parse_empty_port(self):
        parsed = urlutils.URL.from_string("http://example.com:/one")
        self.assertEqual("http", parsed.scheme)
        self.assertIs(None, parsed.user)
        self.assertIs(None, parsed.password)
        self.assertEqual("example.com", parsed.host)
        self.assertIs(None, parsed.port)
        self.assertEqual("/one", parsed.path)


class TestFileRelpath(TestCase):
    # GZ 2011-11-18: A way to override all path handling functions to one
    #                platform or another for testing would be nice.

    def _with_posix_paths(self):
        self.overrideAttr(
            urlutils, "local_path_from_url", urlutils._posix_local_path_from_url
        )
        self.overrideAttr(urlutils, "MIN_ABS_FILEURL_LENGTH", len("file:///"))
        self.overrideAttr(osutils, "normpath", osutils._posix_normpath)
        self.overrideAttr(osutils, "abspath", osutils.posixpath.abspath)
        self.overrideAttr(osutils, "normpath", osutils._posix_normpath)
        self.overrideAttr(osutils, "pathjoin", osutils.posixpath.join)
        self.overrideAttr(osutils, "split", osutils.posixpath.split)
        self.overrideAttr(osutils, "MIN_ABS_PATHLENGTH", 1)

    def _with_win32_paths(self):
        self.overrideAttr(
            urlutils, "local_path_from_url", urlutils._win32_local_path_from_url
        )
        self.overrideAttr(
            urlutils, "MIN_ABS_FILEURL_LENGTH", urlutils.WIN32_MIN_ABS_FILEURL_LENGTH
        )
        self.overrideAttr(osutils, "abspath", osutils._win32_abspath)
        self.overrideAttr(osutils, "normpath", osutils._win32_normpath)
        self.overrideAttr(osutils, "pathjoin", osutils._win32_pathjoin)
        self.overrideAttr(osutils, "split", osutils.ntpath.split)
        self.overrideAttr(osutils, "MIN_ABS_PATHLENGTH", 3)

    def test_same_url_posix(self):
        self._with_posix_paths()
        self.assertEqual("", urlutils.file_relpath("file:///a", "file:///a"))
        self.assertEqual("", urlutils.file_relpath("file:///a", "file:///a/"))
        self.assertEqual("", urlutils.file_relpath("file:///a/", "file:///a"))

    def test_same_url_win32(self):
        self._with_win32_paths()
        self.assertEqual("", urlutils.file_relpath("file:///A:/", "file:///A:/"))
        self.assertEqual("", urlutils.file_relpath("file:///A|/", "file:///A:/"))
        self.assertEqual("", urlutils.file_relpath("file:///A:/b/", "file:///A:/b/"))
        self.assertEqual("", urlutils.file_relpath("file:///A:/b", "file:///A:/b/"))
        self.assertEqual("", urlutils.file_relpath("file:///A:/b/", "file:///A:/b"))

    def test_child_posix(self):
        self._with_posix_paths()
        self.assertEqual("b", urlutils.file_relpath("file:///a", "file:///a/b"))
        self.assertEqual("b", urlutils.file_relpath("file:///a/", "file:///a/b"))
        self.assertEqual("b/c", urlutils.file_relpath("file:///a", "file:///a/b/c"))

    def test_child_win32(self):
        self._with_win32_paths()
        self.assertEqual("b", urlutils.file_relpath("file:///A:/", "file:///A:/b"))
        self.assertEqual("b", urlutils.file_relpath("file:///A|/", "file:///A:/b"))
        self.assertEqual("c", urlutils.file_relpath("file:///A:/b", "file:///A:/b/c"))
        self.assertEqual("c", urlutils.file_relpath("file:///A:/b/", "file:///A:/b/c"))
        self.assertEqual(
            "c/d", urlutils.file_relpath("file:///A:/b", "file:///A:/b/c/d")
        )

    def test_sibling_posix(self):
        self._with_posix_paths()
        self.assertRaises(
            PathNotChild, urlutils.file_relpath, "file:///a/b", "file:///a/c"
        )
        self.assertRaises(
            PathNotChild, urlutils.file_relpath, "file:///a/b/", "file:///a/c"
        )
        self.assertRaises(
            PathNotChild, urlutils.file_relpath, "file:///a/b/", "file:///a/c/"
        )

    def test_sibling_win32(self):
        self._with_win32_paths()
        self.assertRaises(
            PathNotChild, urlutils.file_relpath, "file:///A:/b", "file:///A:/c"
        )
        self.assertRaises(
            PathNotChild, urlutils.file_relpath, "file:///A:/b/", "file:///A:/c"
        )
        self.assertRaises(
            PathNotChild, urlutils.file_relpath, "file:///A:/b/", "file:///A:/c/"
        )

    def test_parent_posix(self):
        self._with_posix_paths()
        self.assertRaises(
            PathNotChild, urlutils.file_relpath, "file:///a/b", "file:///a"
        )
        self.assertRaises(
            PathNotChild, urlutils.file_relpath, "file:///a/b", "file:///a/"
        )

    def test_parent_win32(self):
        self._with_win32_paths()
        self.assertRaises(
            PathNotChild, urlutils.file_relpath, "file:///A:/b", "file:///A:/"
        )
        self.assertRaises(
            PathNotChild, urlutils.file_relpath, "file:///A:/b/c", "file:///A:/b"
        )


class QuoteTests(TestCase):
    def test_quote(self):
        self.assertEqual("abc%20def", urlutils.quote("abc def"))
        self.assertEqual("abc%2Fdef", urlutils.quote("abc/def", safe=""))
        self.assertEqual("abc/def", urlutils.quote("abc/def", safe="/"))

    def test_quote_tildes(self):
        self.assertEqual("~foo", urlutils.quote("~foo"))
        self.assertEqual("~foo", urlutils.quote("~foo", safe="/~"))

    def test_unquote(self):
        self.assertEqual("%", urlutils.unquote("%25"))
        self.assertEqual("\xe5", urlutils.unquote("%C3%A5"))
        self.assertEqual("\xe5", urlutils.unquote("\xe5"))

    def test_unquote_to_bytes(self):
        self.assertEqual(b"%", urlutils.unquote_to_bytes("%25"))
        self.assertEqual(b"\xc3\xa5", urlutils.unquote_to_bytes("%C3%A5"))



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the BranchMapper methods."""

from .... import tests
from .. import branch_mapper
from . import FastimportFeature


class TestBranchMapper(tests.TestCase):
    _test_needs_features = [FastimportFeature]

    def test_git_to_bzr(self):
        m = branch_mapper.BranchMapper()
        for git, bzr in {
            b"refs/heads/master": "trunk",
            b"refs/heads/foo": "foo",
            b"refs/tags/master": "trunk.tag",
            b"refs/tags/foo": "foo.tag",
            b"refs/remotes/origin/master": "trunk.remote",
            b"refs/remotes/origin/foo": "foo.remote",
        }.items():
            self.assertEqual(m.git_to_bzr(git), bzr)

    def test_git_to_bzr_with_slashes(self):
        m = branch_mapper.BranchMapper()
        for git, bzr in {
            b"refs/heads/master/slave": "master/slave",
            b"refs/heads/foo/bar": "foo/bar",
            b"refs/tags/master/slave": "master/slave.tag",
            b"refs/tags/foo/bar": "foo/bar.tag",
            b"refs/remotes/origin/master/slave": "master/slave.remote",
            b"refs/remotes/origin/foo/bar": "foo/bar.remote",
        }.items():
            self.assertEqual(m.git_to_bzr(git), bzr)

    def test_git_to_bzr_for_trunk(self):
        # As 'master' in git is mapped to trunk in bzr, we need to handle
        # 'trunk' in git in a sensible way.
        m = branch_mapper.BranchMapper()
        for git, bzr in {
            b"refs/heads/trunk": "git-trunk",
            b"refs/tags/trunk": "git-trunk.tag",
            b"refs/remotes/origin/trunk": "git-trunk.remote",
            b"refs/heads/git-trunk": "git-git-trunk",
            b"refs/tags/git-trunk": "git-git-trunk.tag",
            b"refs/remotes/origin/git-trunk": "git-git-trunk.remote",
        }.items():
            self.assertEqual(m.git_to_bzr(git), bzr)



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that all Tree's implement .annotate_iter()."""

from breezy.tests.per_tree import TestCaseWithTree


class TestAnnotate(TestCaseWithTree):
    def get_simple_tree(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/one", b"first\ncontent\n")])
        tree.add(["one"])
        rev_1 = tree.commit("one")
        self.build_tree_contents([("tree/one", b"second\ncontent\n")])
        rev_2 = tree.commit("two")
        return self._convert_tree(tree), [rev_1, rev_2]

    def get_tree_with_ghost(self):
        tree = self.make_branch_and_tree("tree")
        if not tree.branch.repository._format.supports_ghosts:
            self.skipTest("repository format does not support ghosts")
        self.build_tree_contents([("tree/one", b"first\ncontent\n")])
        tree.add(["one"])
        rev_1 = tree.commit("one")
        tree.set_parent_ids([rev_1, b"ghost-one"])
        self.build_tree_contents([("tree/one", b"second\ncontent\n")])
        rev_2 = tree.commit("two")
        return self._convert_tree(tree), [rev_1, rev_2]

    def test_annotate_simple(self):
        tree, revids = self.get_simple_tree()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(
            [(revids[1], b"second\n"), (revids[0], b"content\n")],
            list(tree.annotate_iter("one")),
        )

    def test_annotate_with_ghost(self):
        tree, revids = self.get_tree_with_ghost()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(
            [(revids[1], b"second\n"), (revids[0], b"content\n")],
            list(tree.annotate_iter("one")),
        )



# Copyright (C) 2006 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black-box tests for 'brz inventory'."""

import os

from breezy.tests import TestCaseWithTransport


class TestInventory(TestCaseWithTransport):
    def setUp(self):
        super().setUp()

        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/c"])

        tree.add(["a", "b", "b/c"], ids=[b"a-id", b"b-id", b"c-id"])
        tree.commit("init", rev_id=b"one")
        self.tree = tree

    def assertInventoryEqual(self, expected, args=None, **kwargs):
        """Test that the output of 'brz inventory' is as expected.

        Any arguments supplied will be passed to run_bzr.
        """
        command = "inventory"
        if args is not None:
            command += " " + args
        out, err = self.run_bzr(command, **kwargs)
        self.assertEqual(expected, out)
        self.assertEqual("", err)

    def test_inventory(self):
        self.assertInventoryEqual("a\nb\nb/c\n")

    def test_inventory_include_root(self):
        self.assertInventoryEqual("\na\nb\nb/c\n", "--include-root")
        self.assertInventoryEqual("b\nb/c\n", "--include-root b")

    def test_inventory_kind(self):
        self.assertInventoryEqual("a\nb/c\n", "--kind file")
        self.assertInventoryEqual("b\n", "--kind directory")

    def test_inventory_show_ids(self):
        expected = "".join(
            ("%-50s %s\n" % (path, file_id))
            for path, file_id in [("a", "a-id"), ("b", "b-id"), ("b/c", "c-id")]
        )
        self.assertInventoryEqual(expected, "--show-ids")

    def test_inventory_specific_files(self):
        self.assertInventoryEqual("a\n", "a")
        self.assertInventoryEqual("b\nb/c\n", "b b/c")
        # 'brz inventory' recurses into subdirectories
        self.assertInventoryEqual("b\nb/c\n", "b")

    def test_inventory_mixed(self):
        """Test that we get expected results when mixing parameters."""
        a_line = "%-50s %s\n" % ("a", "a-id")
        c_line = "%-50s %s\n" % ("b/c", "c-id")

        self.assertInventoryEqual("", "--kind directory a")
        self.assertInventoryEqual(a_line + c_line, "--kind file --show-ids")
        self.assertInventoryEqual(c_line, "--kind file --show-ids b b/c")

    def test_in_subdir(self):
        os.chdir("b")
        # TODO: jam 20060922 Maybe inventory should return the paths as
        #       relative to '.', rather than relative to root

        # a plain 'inventory' returns all files
        self.assertInventoryEqual("a\nb\nb/c\n")
        # But passing '.' will only return paths underneath here
        self.assertInventoryEqual("b\nb/c\n", ".")

    def test_inventory_revision(self):
        self.build_tree(["b/d", "e"])
        self.tree.add(["b/d", "e"], ids=[b"d-id", b"e-id"])
        self.tree.commit("add files")

        self.tree.rename_one("b/d", "d")
        self.tree.commit("rename b/d => d")

        # Passing just -r returns the inventory of that revision
        self.assertInventoryEqual("a\nb\nb/c\n", "-r 1")
        self.assertInventoryEqual("a\nb\nb/c\nb/d\ne\n", "-r 2")

        # Passing a path will lookup the path in the old and current locations
        self.assertInventoryEqual("b/d\n", "-r 2 b/d")
        self.assertInventoryEqual("b/d\n", "-r 2 d")

        self.tree.rename_one("e", "b/e")
        self.tree.commit("rename e => b/e")

        # When supplying just a directory paths that are now,
        # or used to be, in that directory are shown
        self.assertInventoryEqual("b\nb/c\nb/d\ne\n", "-r 2 b")

    def test_missing_file(self):
        self.run_bzr_error(
            [r"Path\(s\) are not versioned: no-such-file"], "inventory no-such-file"
        )
